import pytest
import numpy as np


def test_add_samples_to_existing_column(repo_20_filled_samples2):
    from hangar.records.summarize import status
    repo = repo_20_filled_samples2
    expected = '============ \n'\
               '| Branch: master \n'\
               ' \n'\
               '============ \n'\
               '| ADDED \n'\
               '|---------- \n'\
               '| Schema: 0 \n'\
               '|---------- \n'\
               '| Samples: 20 \n'\
               '|  - "dummy": 20 \n'\
               ' \n'\
               '============ \n'\
               '| DELETED \n'\
               '|---------- \n'\
               '| Schema: 0 \n'\
               '|---------- \n'\
               '| Samples: 0 \n'\
               ' \n'\
               '============ \n'\
               '| MUTATED \n'\
               '|---------- \n'\
               '| Schema: 0 \n'\
               '|---------- \n'\
               '| Samples: 0 \n'\
               ' \n'
    dummyData = np.arange(50).astype(np.int64)
    co2 = repo.checkout(write=True)
    for idx in range(10, 20):
        dummyData[:] = idx
        co2.columns['dummy'][str(idx)] = dummyData
        co2.columns['dummy'][idx] = dummyData
    df = co2.diff.staged()
    co2.close()
    assert status(repo._env.hashenv, 'master', df.diff).getvalue() == expected


def test_mutate_sample_values(repo_20_filled_samples2):
    from hangar.records.summarize import status
    repo = repo = repo_20_filled_samples2
    expected = '============ \n'\
               '| Branch: master \n'\
               ' \n'\
               '============ \n'\
               '| ADDED \n'\
               '|---------- \n'\
               '| Schema: 0 \n'\
               '|---------- \n'\
               '| Samples: 0 \n'\
               ' \n'\
               '============ \n'\
               '| DELETED \n'\
               '|---------- \n'\
               '| Schema: 0 \n'\
               '|---------- \n'\
               '| Samples: 0 \n'\
               ' \n'\
               '============ \n'\
               '| MUTATED \n'\
               '|---------- \n'\
               '| Schema: 0 \n'\
               '|---------- \n'\
               '| Samples: 5 \n'\
               '|  - "dummy": 5 \n'\
               ' \n'

    dummyData = np.arange(50).astype(np.int64)
    co2 = repo.checkout(write=True)
    for idx in range(5, 10):
        dummyData[:] = idx + 10
        co2.columns['dummy'][idx] = dummyData
    df = co2.diff.staged()
    co2.close()
    assert status(repo._env.hashenv, 'master', df.diff).getvalue() == expected


def test_delete_samples(repo_20_filled_samples2):
    from hangar.records.summarize import status
    repo = repo_20_filled_samples2
    expected = '============ \n'\
               '| Branch: master \n'\
               ' \n'\
               '============ \n'\
               '| ADDED \n'\
               '|---------- \n'\
               '| Schema: 0 \n'\
               '|---------- \n'\
               '| Samples: 0 \n'\
               ' \n'\
               '============ \n'\
               '| DELETED \n'\
               '|---------- \n'\
               '| Schema: 0 \n'\
               '|---------- \n'\
               '| Samples: 5 \n'\
               '|  - "dummy": 5 \n'\
               ' \n'\
               '============ \n'\
               '| MUTATED \n'\
               '|---------- \n'\
               '| Schema: 0 \n'\
               '|---------- \n'\
               '| Samples: 0 \n'\
               ' \n'

    co2 = repo.checkout(write=True)
    for idx in range(5, 10):
        del co2.columns['dummy'][idx]
    df = co2.diff.staged()
    co2.close()
    assert status(repo._env.hashenv, 'master', df.diff).getvalue() == expected


def test_add_new_column_schema_and_samples(repo_20_filled_samples2):
    from hangar.records.summarize import status
    repo = repo_20_filled_samples2
    expected = (
        '============ \n'
        '| Branch: master \n'
        ' \n'
        '============ \n'
        '| ADDED \n'
        '|---------- \n'
        '| Schema: 1 \n'
        '|  - "new_aset": \n'
        '|       digest="1=555a833b66ab" \n'
        '|       column_layout: flat \n'
        '|       column_type: ndarray \n'
        '|       schema_hasher_tcode: 1 \n'
        '|       data_hasher_tcode: 0 \n'
        '|       schema_type: fixed_shape \n'
        '|       shape: (10, 10) \n'
        '|       dtype: float32 \n'
        '|       backend: 01 \n'
        '|       backend_options: {\'complib\': \'blosc:lz4hc\', \'complevel\': 5, \'shuffle\': \'byte\'} \n'
        '|---------- \n'
        '| Samples: 5 \n'
        '|  - "new_aset": 5 \n'
        ' \n'
        '============ \n'
        '| DELETED \n'
        '|---------- \n'
        '| Schema: 0 \n'
        '|---------- \n'
        '| Samples: 0 \n'
        ' \n'
        '============ \n'
        '| MUTATED \n'
        '|---------- \n'
        '| Schema: 0 \n'
        '|---------- \n'
        '| Samples: 0 \n'
        ' \n'
    )
    co2 = repo.checkout(write=True)
    co2.add_ndarray_column('new_aset', shape=(10, 10), dtype=np.float32)
    for idx in range(5):
        dummyData = np.random.randn(10, 10).astype(np.float32)
        co2.columns['new_aset'][idx] = dummyData
    df = co2.diff.staged()
    co2.close()
    result = status(repo._env.hashenv, 'master', df.diff).getvalue()
    assert result == expected


def test_add_new_column_schema_and_sample_and_delete_old_column(repo_20_filled_samples2):
    from hangar.records.summarize import status
    repo = repo_20_filled_samples2
    expected = (
        '============ \n'
        '| Branch: master \n'
        ' \n'
        '============ \n'
        '| ADDED \n'
        '|---------- \n'
        '| Schema: 1 \n'
        '|  - "new_aset": \n'
        '|       digest="1=555a833b66ab" \n'
        '|       column_layout: flat \n'
        '|       column_type: ndarray \n'
        '|       schema_hasher_tcode: 1 \n'
        '|       data_hasher_tcode: 0 \n'
        '|       schema_type: fixed_shape \n'
        '|       shape: (10, 10) \n'
        '|       dtype: float32 \n'
        '|       backend: 01 \n'
        '|       backend_options: {\'complib\': \'blosc:lz4hc\', \'complevel\': 5, \'shuffle\': \'byte\'} \n'
        '|---------- \n'
        '| Samples: 5 \n'
        '|  - "new_aset": 5 \n'
        ' \n'
        '============ \n'
        '| DELETED \n'
        '|---------- \n'
        '| Schema: 1 \n'
        '|  - "dummy": \n'
        '|       digest="1=18599cd5ea25" \n'
        '|       column_layout: flat \n'
        '|       column_type: ndarray \n'
        '|       schema_hasher_tcode: 1 \n'
        '|       data_hasher_tcode: 0 \n'
        '|       schema_type: fixed_shape \n'
        '|       shape: (50,) \n'
        '|       dtype: int64 \n'
        '|       backend: 10 \n'
        '|       backend_options: {} \n'
        '|---------- \n'
        '| Samples: 10 \n'
        '|  - "dummy": 10 \n'
        ' \n'
        '============ \n'
        '| MUTATED \n'
        '|---------- \n'
        '| Schema: 0 \n'
        '|---------- \n'
        '| Samples: 0 \n'
        ' \n'
    )
    co2 = repo.checkout(write=True)
    new = co2.add_ndarray_column('new_aset', shape=(10, 10), dtype=np.float32)
    for idx in range(5):
               dummyData = np.random.randn(10, 10).astype(np.float32)
               co2.columns['new_aset'][idx] = dummyData
    del co2.columns['dummy']
    df = co2.diff.staged()
    co2.close()
    result = status(repo._env.hashenv, 'master', df.diff).getvalue()
    assert result == expected


def test_add_new_schema_and_samples_and_change_old_backend(repo_20_filled_samples2):
    from hangar.records.summarize import status
    repo = repo_20_filled_samples2
    expected = (
        '============ \n'
        '| Branch: master \n'
        ' \n'
        '============ \n'
        '| ADDED \n'
        '|---------- \n'
        '| Schema: 1 \n'
        '|  - "new_aset": \n'
        '|       digest="1=555a833b66ab" \n'
        '|       column_layout: flat \n'
        '|       column_type: ndarray \n'
        '|       schema_hasher_tcode: 1 \n'
        '|       data_hasher_tcode: 0 \n'
        '|       schema_type: fixed_shape \n'
        '|       shape: (10, 10) \n'
        '|       dtype: float32 \n'
        '|       backend: 01 \n'
        '|       backend_options: {\'complib\': \'blosc:lz4hc\', \'complevel\': 5, \'shuffle\': \'byte\'} \n'
        '|---------- \n'
        '| Samples: 5 \n'
        '|  - "new_aset": 5 \n'
        ' \n'
        '============ \n'
        '| DELETED \n'
        '|---------- \n'
        '| Schema: 0 \n'
        '|---------- \n'
        '| Samples: 0 \n'
        ' \n'
        '============ \n'
        '| MUTATED \n'
        '|---------- \n'
        '| Schema: 1 \n'
        '|  - "dummy": \n'
        '|       digest="1=5d6cc8241705" \n'
        '|       column_layout: flat \n'
        '|       column_type: ndarray \n'
        '|       schema_hasher_tcode: 1 \n'
        '|       data_hasher_tcode: 0 \n'
        '|       schema_type: fixed_shape \n'
        '|       shape: (50,) \n'
        '|       dtype: int64 \n'
        '|       backend: 00 \n'
        '|       backend_options: {\'complib\': \'blosc:lz4hc\', \'complevel\': 5, \'shuffle\': \'byte\'} \n'
        '|---------- \n'
        '| Samples: 5 \n'
        '|  - "dummy": 5 \n'
        ' \n'
    )
    co2 = repo.checkout(write=True)
    co2.columns['dummy'].change_backend('00')
    co2.add_ndarray_column('new_aset', shape=(10, 10), dtype=np.float32)
    for idx in range(5):
        dummyData = np.random.randn(10, 10).astype(np.float32)
        co2.columns['new_aset'][idx] = dummyData
        co2.columns['dummy'][idx] = np.arange(50).astype(np.int64) + idx
    df = co2.diff.staged()
    co2.close()
    result = status(repo._env.hashenv, 'master', df.diff).getvalue()
    assert result == expected



from pathlib import Path

import lmdb

from .parsing import (
    repo_version_db_key,
    repo_version_db_val_from_raw_val,
    repo_version_raw_spec_from_raw_string,
    repo_version_raw_val_from_db_val,
)
from .._version import Version
from ..constants import LMDB_SETTINGS, LMDB_BRANCH_NAME
from ..txnctx import TxnRegister
from ..utils import pairwise


def set_repository_software_version(branchenv: lmdb.Environment,
                                    ver_str: str,
                                    *,
                                    overwrite: bool = False) -> bool:
    """Write the repository software version to a particular value

    Parameters
    ----------
    branchenv : lmdb.Environment
        db where the head, branch, and version specs are stored
    ver_str : str
        semantic version style string representing version (ie. "0.1.0",
        "1.2.1", etc)
    overwrite : bool, optional
        If True, replace current value with new value; If False, do not
        overwrite if this key exists, by default False

    Returns
    -------
    bool
        True if successful, False otherwise
    """
    versionKey = repo_version_db_key()
    ver_spec = repo_version_raw_spec_from_raw_string(v_str=ver_str)
    versionVal = repo_version_db_val_from_raw_val(v_spec=ver_spec)
    branchTxn = TxnRegister().begin_writer_txn(branchenv)
    try:
        success = branchTxn.put(versionKey, versionVal, overwrite=overwrite)
    finally:
        TxnRegister().commit_writer_txn(branchenv)
    return success


def get_repository_software_version_spec(branchenv: lmdb.Environment) -> Version:
    """Get the repository version specification tuple.

    Parameters
    ----------
    branchenv : lmdb.Environment
        db where the head, branch, and version specs are stored

    Returns
    -------
    Version
        This class abstracts handling of a project’s versions. A Version
        instance is comparison aware and can be compared and sorted using the
        standard Python interfaces.

    Raises
    ------
    KeyError
        If no version key is set for the repository
    """
    versionKey = repo_version_db_key()
    branchTxn = TxnRegister().begin_reader_txn(branchenv)
    try:
        versionVal = branchTxn.get(versionKey, default=False)
    finally:
        TxnRegister().abort_reader_txn(branchenv)

    if versionVal is False:
        raise KeyError('No version string is set for the repository')
    else:
        version_val = repo_version_raw_val_from_db_val(versionVal)
        return version_val


"""
Initial checking of repository versions
---------------------------------------
"""


def startup_check_repo_version(repo_path: Path) -> Version:
    """Determine repo version without having to have Environments ctx opened.

    Parameters
    ----------
    repo_path : Path
        path to the repository directory on disk

    Returns
    -------
    Version
        This class abstracts handling of a project’s versions. A Version
        instance is comparison aware and can be compared and sorted using the
        standard Python interfaces.

    Raises
    ------
    RuntimeError
        If for whatever reason, the branch file does not exist on disk.
        Execution should not reach this point.
    """
    brch_fp = repo_path.joinpath(LMDB_BRANCH_NAME)
    if not brch_fp.is_file():
        msg = f'Hangar Internal Error, startup_check_repo_version did not find '\
              f'brch db at: {brch_fp}. Execution should never reach this point. '\
              f'Please report this error to Hangar developers.'
        raise RuntimeError(msg)

    branchenv = lmdb.open(path=str(brch_fp), readonly=True, create=False, **LMDB_SETTINGS)
    spec = get_repository_software_version_spec(branchenv=branchenv)
    branchenv.close()
    return spec


incompatible_changes_after = [
    Version('0.2.0'),
    Version('0.3.0'),
    Version('0.4.0'),
    Version('0.5.0.dev0'),
    Version('0.5.0.dev1'),
]


def is_repo_software_version_compatible(repo_v: Version, curr_v: Version) -> bool:
    """Determine if the repo on disk and the current Hangar versions iscompatible.

    Parameters
    ----------
    repo_v : Version
        repository software writtern version.
    curr_v : Version
        currently active software version specification

    Returns
    -------
    bool
        True if compatible, False if not.
    """
    for start, end in pairwise(incompatible_changes_after):
        if (repo_v >= start) and (repo_v < end):
            if (curr_v < start) or (curr_v >= end):
                return False
            elif (curr_v >= start) and (curr_v < end):
                return True
    if (repo_v >= end) and (curr_v < end):
        return False
    return True



__version__ = '0.5.0'
__all__ = ('make_torch_dataset', 'make_tf_dataset', 'Repository')

from .repository import Repository


def __getattr__(name):  # lazy loader
    if name == 'make_tf_dataset':
        from .dataloaders import make_tf_dataset
        return make_tf_dataset

    if name == 'make_torch_dataset':
        from .dataloaders import make_torch_dataset
        return make_torch_dataset

    raise AttributeError(f"module {__name__} has no attribute {name}")



import pytest

import numpy as np


@pytest.fixture()
def diverse_repo(repo):
    co = repo.checkout(write=True)
    co.add_ndarray_column('test', prototype=np.arange(10))
    co.add_str_column('test_meta')
    co.columns['test'][0] = np.arange(10)
    co.columns['test'][1] = np.arange(10) + 1
    co.columns['test'][2] = np.arange(10) + 2
    co.columns['test'][3] = np.arange(10) + 3
    co.columns['test'][4] = np.arange(10) + 4
    co['test_meta']['hi'] = 'foo'
    co['test_meta']['aea'] = 'eeae'
    co.commit('hello world')

    sample_trimg = np.arange(50).reshape(5, 10).astype(np.uint8)
    sample_trlabel = np.array([0], dtype=np.int64)
    sample_vimg = np.zeros(50).reshape(5, 10).astype(np.uint16)
    sample_vlabel = np.array([1], dtype=np.int32)

    co.close()
    repo.create_branch('dev')
    co = repo.checkout(write=True, branch='dev')
    dset_trlabels = co.add_ndarray_column(name='train_labels', prototype=sample_trlabel)
    dset_trimgs = co.add_ndarray_column('train_images', prototype=sample_trimg, backend='01')
    dset_trlabels[0] = sample_trlabel
    dset_trlabels[1] = sample_trlabel + 1
    dset_trlabels[2] = sample_trlabel + 2
    dset_trimgs[0] = sample_trimg
    dset_trimgs[1] = sample_trimg + 1
    dset_trimgs[2] = sample_trimg + 2
    co.commit('second on dev')
    co.close()

    co = repo.checkout(write=True, branch='master')
    dset_vimgs = co.add_ndarray_column('valid_images', prototype=sample_vimg)
    dset_vlabels = co.add_ndarray_column('valid_labels', prototype=sample_vlabel)
    dset_vlabels[0] = sample_vlabel
    dset_vlabels[1] = sample_vlabel + 1
    dset_vlabels[2] = sample_vlabel + 2
    dset_vimgs[0] = sample_vimg
    dset_vimgs[1] = sample_vimg + 1
    dset_vimgs[2] = sample_vimg + 2
    co['test_meta']['second'] = 'on master now'
    co.commit('second on master')
    co.close()

    base = repo.merge('merge commit', 'master', 'dev')
    repo.create_branch('newbranch', base_commit=base)
    co = repo.checkout(write=True, branch='master')
    co['test_meta']['newmeta'] = 'wow'
    co.commit('on master after merge')
    co.close()

    co = repo.checkout(write=True, branch='newbranch')
    ds_trimgs = co.columns['train_images']
    ds_trlabels = co.columns['train_labels']
    ds_trlabels[3] = sample_trlabel + 3
    ds_trlabels[4] = sample_trlabel + 4
    ds_trlabels[5] = sample_trlabel + 5
    ds_trimgs[3] = sample_trimg + 3
    ds_trimgs[4] = sample_trimg + 4
    ds_trimgs[5] = sample_trimg + 5
    co.commit('on newdev after merge')
    co.close()

    base = repo.merge('new merge commit', 'master', 'newbranch')
    return repo


def test_verify_correct(diverse_repo):
    assert diverse_repo.verify_repo_integrity() is True


class TestVerifyCommitRefDigests(object):

    def test_remove_array_digest_is_caught(self, diverse_repo):
        from hangar.records import hashs
        from hangar.diagnostics.integrity import _verify_commit_ref_digests_exist

        hq = hashs.HashQuery(diverse_repo._env.hashenv)
        keys_to_remove = list(hq.gen_all_hash_keys_db())

        for key_removed in keys_to_remove:
            with diverse_repo._env.hashenv.begin(write=True) as txn:
                val_removed = txn.get(key_removed)
                txn.delete(key_removed)

            with pytest.raises(RuntimeError):
                _verify_commit_ref_digests_exist(hashenv=diverse_repo._env.hashenv,
                                                 refenv=diverse_repo._env.refenv)

            with diverse_repo._env.hashenv.begin(write=True) as txn:
                txn.put(key_removed, val_removed)

    def test_remove_schema_digest_is_caught(self, diverse_repo):
        from hangar.records import hashs
        from hangar.diagnostics.integrity import _verify_commit_ref_digests_exist

        hq = hashs.HashQuery(diverse_repo._env.hashenv)
        keys_to_remove = list(hq.gen_all_schema_keys_db())
        for key_removed in keys_to_remove:
            with diverse_repo._env.hashenv.begin(write=True) as txn:
                val_removed = txn.get(key_removed)
                txn.delete(key_removed)

            with pytest.raises(RuntimeError):
                _verify_commit_ref_digests_exist(hashenv=diverse_repo._env.hashenv,
                                                 refenv=diverse_repo._env.refenv)

            with diverse_repo._env.hashenv.begin(write=True) as txn:
                txn.put(key_removed, val_removed)


class TestVerifyCommitTree(object):

    def test_parent_ref_digest_of_cmt_does_not_exist(self, diverse_repo):
        from hangar.diagnostics.integrity import _verify_commit_tree_integrity
        from hangar.records.parsing import commit_parent_db_key_from_raw_key

        repo = diverse_repo
        history = repo.log(return_contents=True)
        all_commits = history['order']
        for cmt in all_commits:
            parentKey = commit_parent_db_key_from_raw_key(cmt)
            with repo._env.refenv.begin(write=True) as txn:
                parentVal = txn.get(parentKey)
                txn.delete(parentKey)

            with pytest.raises(RuntimeError, match='Data corruption detected for parent ref'):
                _verify_commit_tree_integrity(repo._env.refenv)

            with repo._env.refenv.begin(write=True) as txn:
                txn.put(parentKey, parentVal)

    def test_parent_ref_references_nonexisting_commits(self, diverse_repo):
        from hangar.diagnostics.integrity import _verify_commit_tree_integrity
        from hangar.records.parsing import commit_parent_db_key_from_raw_key
        from hangar.records.parsing import commit_parent_raw_val_from_db_val
        from hangar.records.parsing import commit_parent_db_val_from_raw_val

        repo = diverse_repo
        history = repo.log(return_contents=True)
        all_commits = history['order']
        for cmt in all_commits:
            parentKey = commit_parent_db_key_from_raw_key(cmt)
            with repo._env.refenv.begin(write=True) as txn:
                parentVal = txn.get(parentKey)
                parent_raw = commit_parent_raw_val_from_db_val(parentVal)
                parent = parent_raw.ancestor_spec

                if parent.dev_ancestor:
                    modifiedVal = commit_parent_db_val_from_raw_val(
                        master_ancestor=parent.master_ancestor,
                        dev_ancestor='corrupt',
                        is_merge_commit=parent.is_merge_commit)
                elif parent.master_ancestor:
                    modifiedVal = commit_parent_db_val_from_raw_val(
                        master_ancestor='corrupt',
                        dev_ancestor=parent.dev_ancestor,
                        is_merge_commit=parent.is_merge_commit)
                else:
                    continue

                txn.put(parentKey, modifiedVal.raw, overwrite=True)

            with pytest.raises(RuntimeError, match='Data corruption detected in commit tree'):
                _verify_commit_tree_integrity(repo._env.refenv)

            with repo._env.refenv.begin(write=True) as txn:
                txn.put(parentKey, parentVal)

    def test_parent_ref_has_two_initial_commits(self, diverse_repo):
        from hangar.diagnostics.integrity import _verify_commit_tree_integrity
        from hangar.records.parsing import commit_parent_db_key_from_raw_key

        repo = diverse_repo
        repo = diverse_repo
        history = repo.log(return_contents=True)
        all_commits = history['order']
        initial_commit = all_commits[-1]
        for cmt in all_commits:
            if cmt == initial_commit:
                continue

            parentKey = commit_parent_db_key_from_raw_key(cmt)
            with repo._env.refenv.begin(write=True) as txn:
                parentVal = txn.get(parentKey)
                txn.put(parentKey, b'', overwrite=True)

            with pytest.raises(RuntimeError, match='Commit tree integrity compromised. Multiple "initial"'):
                _verify_commit_tree_integrity(repo._env.refenv)

            with repo._env.refenv.begin(write=True) as txn:
                txn.put(parentKey, parentVal, overwrite=True)


class TestBranchIntegrity(object):

    def test_atleast_one_branch_exists(self, diverse_repo):
        from hangar.records.heads import get_branch_names
        from hangar.records.parsing import repo_branch_head_db_key_from_raw_key
        from hangar.diagnostics.integrity import _verify_branch_integrity

        branch_names = get_branch_names(diverse_repo._env.branchenv)
        with diverse_repo._env.branchenv.begin(write=True) as txn:
            for bname in branch_names:
                branchKey = repo_branch_head_db_key_from_raw_key(bname)
                txn.delete(branchKey)

        with pytest.raises(
                RuntimeError,
                match='Branch map compromised. Repo must contain atleast one branch'
        ):
            _verify_branch_integrity(diverse_repo._env.branchenv, diverse_repo._env.refenv)

    def test_branch_name_head_commit_digests_exist(self, diverse_repo):
        from hangar.records.heads import get_branch_names, get_branch_head_commit
        from hangar.records.parsing import commit_ref_db_key_from_raw_key
        from hangar.records.parsing import commit_parent_db_key_from_raw_key
        from hangar.records.parsing import commit_spec_db_key_from_raw_key
        from hangar.diagnostics.integrity import _verify_branch_integrity

        branch_names = get_branch_names(diverse_repo._env.branchenv)
        for bname in branch_names:
            bhead = get_branch_head_commit(diverse_repo._env.branchenv, branch_name=bname)
            with diverse_repo._env.refenv.begin(write=True) as txn:
                cmtRefKey = commit_ref_db_key_from_raw_key(bhead)
                cmtSpecKey = commit_spec_db_key_from_raw_key(bhead)
                cmtParentKey = commit_parent_db_key_from_raw_key(bhead)

                cmtRefVal = txn.get(cmtRefKey)
                cmtSpecVal = txn.get(cmtSpecKey)
                cmtParentVal = txn.get(cmtParentKey)

                txn.delete(cmtRefKey)
                txn.delete(cmtSpecKey)
                txn.delete(cmtParentKey)

            with pytest.raises(RuntimeError, match='Branch commit map compromised. Branch name'):
                _verify_branch_integrity(diverse_repo._env.branchenv, diverse_repo._env.refenv)

            with diverse_repo._env.refenv.begin(write=True) as txn:
                txn.put(cmtRefKey, cmtRefVal)
                txn.put(cmtSpecKey, cmtSpecVal)
                txn.put(cmtParentKey, cmtParentVal)

    def test_staging_head_branch_name_exists(self, diverse_repo):
        from hangar.records.heads import get_staging_branch_head
        from hangar.records.parsing import repo_branch_head_db_key_from_raw_key
        from hangar.diagnostics.integrity import _verify_branch_integrity

        bname = get_staging_branch_head(diverse_repo._env.branchenv)
        with diverse_repo._env.branchenv.begin(write=True) as txn:
            branchKey = repo_branch_head_db_key_from_raw_key(bname)
            txn.delete(branchKey)

        with pytest.raises(
                RuntimeError,
                match='Brach commit map compromised. Staging head refers to branch name'
        ):
            _verify_branch_integrity(diverse_repo._env.branchenv, diverse_repo._env.refenv)


def test_data_digest_modification_is_caught(diverse_repo):
    from hangar.records import hashs
    from hangar.diagnostics.integrity import _verify_column_integrity

    hq = hashs.HashQuery(diverse_repo._env.hashenv)
    keys_to_replace = list(hq.gen_all_hash_keys_db())
    replacer_key = keys_to_replace.pop()
    for kreplaced in keys_to_replace:
        with diverse_repo._env.hashenv.begin(write=True) as txn:
            replacer_val = txn.get(replacer_key)
            vreplaced = txn.get(kreplaced)
            txn.put(kreplaced, replacer_val)

        with pytest.raises(RuntimeError):
            _verify_column_integrity(hashenv=diverse_repo._env.hashenv,
                                     repo_path=diverse_repo._env.repo_path)

        with diverse_repo._env.hashenv.begin(write=True) as txn:
            txn.put(kreplaced, vreplaced)


def test_data_digest_remote_location_warns(diverse_repo):
    from hangar.records import hashs
    from hangar.diagnostics.integrity import _verify_column_integrity

    hq = hashs.HashQuery(diverse_repo._env.hashenv)
    replace_key = list(hq.gen_all_hash_keys_db())[0]
    with diverse_repo._env.hashenv.begin(write=True) as txn:
        txn.put(replace_key, b'50:ekaearar')

    with pytest.warns(RuntimeWarning, match='Can not verify integrity of partially fetched array'):
        _verify_column_integrity(hashenv=diverse_repo._env.hashenv,
                                 repo_path=diverse_repo._env.repo_path)


def test_schema_digest_modification_is_caught(diverse_repo):
    from hangar.records import hashs
    from hangar.diagnostics.integrity import _verify_schema_integrity

    hq = hashs.HashQuery(diverse_repo._env.hashenv)
    keys_to_replace = list(hq.gen_all_schema_keys_db())
    replacer_key = keys_to_replace.pop()
    for kreplaced in keys_to_replace:
        with diverse_repo._env.hashenv.begin(write=True) as txn:
            replacer_val = txn.get(replacer_key)
            vreplaced = txn.get(kreplaced)
            txn.put(kreplaced, replacer_val)

        with pytest.raises(RuntimeError, match='Data corruption detected for schema. Expected digest'):
            _verify_schema_integrity(hashenv=diverse_repo._env.hashenv)

        with diverse_repo._env.hashenv.begin(write=True) as txn:
            txn.put(kreplaced, vreplaced)



from itertools import starmap
from typing import Iterable, List, NamedTuple, Set, Tuple, Union

import lmdb

from .records import (
    dynamic_layout_data_record_from_db_key,
    schema_column_record_from_db_key,
    data_record_digest_val_from_db_val,
    ColumnSchemaKey,
    FlatColumnDataKey,
    NestedColumnDataKey,
)
from .records.commiting import (
    check_commit_hash_in_history,
    get_commit_ancestors_graph,
    get_commit_ref,
    get_commit_spec,
    tmp_cmt_env,
)
from .records.heads import get_branch_head_commit, get_branch_names
from .records.queries import RecordQuery
from .txnctx import TxnRegister

# ------------------------- Differ Types --------------------------------------


class HistoryDiffStruct(NamedTuple):
    masterHEAD: str
    devHEAD: str
    ancestorHEAD: str
    canFF: bool


class Changes(NamedTuple):
    schema: dict
    samples: tuple


class DiffOutDB(NamedTuple):
    added: Set[Tuple[bytes, bytes]]
    deleted: Set[Tuple[bytes, bytes]]
    mutated: Set[Tuple[bytes, bytes]]


class DiffOut(NamedTuple):
    added: Changes
    deleted: Changes
    mutated: Changes


ConflictKeys = Union[str, FlatColumnDataKey, NestedColumnDataKey, ColumnSchemaKey]


class Conflicts(NamedTuple):
    """Four types of conflicts are accessible through this object.

    Attributes
    ----------
    t1
        Addition of key in master AND dev with different values.
    t21
        Removed key in master, mutated value in dev.
    t22
        Removed key in dev, mutated value in master.
    t3
        Mutated key in both master AND dev to different values.
    conflict
        Bool indicating if any type of conflict is present.
    """
    t1: Iterable[ConflictKeys]
    t21: Iterable[ConflictKeys]
    t22: Iterable[ConflictKeys]
    t3: Iterable[ConflictKeys]
    conflict: bool


class DiffAndConflictsDB(NamedTuple):
    diff: DiffOutDB
    conflict: Conflicts


class DiffAndConflicts(NamedTuple):
    diff: DiffOut
    conflict: Conflicts


# ------------------------------- Differ Methods ------------------------------


def diff_envs(base_env: lmdb.Environment, head_env: lmdb.Environment, ) -> DiffOutDB:
    """Main diff algorithm to determine changes between unpacked lmdb environments.

    Parameters
    ----------
    base_env : lmdb.Environment
        starting point to calculate changes from
    head_env : lmdb.Environment
        some commit which should be compared to BASE

    Returns
    -------
    DiffOutDB
        iterable of db formatted key/value pairs for `added`, `deleted`,
        `mutated` fields
    """
    added, deleted, mutated = [], [], []

    baseTxn = TxnRegister().begin_reader_txn(base_env)
    headTxn = TxnRegister().begin_reader_txn(head_env)
    baseCur = baseTxn.cursor()
    headCur = headTxn.cursor()
    try:
        moreBase = baseCur.first()
        moreHead = headCur.first()

        while True:
            if moreBase and moreHead:
                bKey, bVal = baseCur.item()
                hKey, hVal = headCur.item()
            elif (not moreBase) and (not moreHead):
                break
            # necessary to avoid deadlock at last items
            elif not moreBase:
                bKey = b'x'
                bVal = b''
                hKey, hVal = headCur.item()
            else:  # (not moreHead)
                hKey = b'x'
                hVal = b''
                bKey, bVal = baseCur.item()

            # inserted
            if bKey > hKey:
                added.append((hKey, hVal))
                moreHead = headCur.next()
                continue
            # deleted
            elif bKey < hKey:
                deleted.append((bKey, bVal))
                moreBase = baseCur.next()
                continue
            # no change
            elif (bKey == hKey) and (bVal == hVal):
                moreBase = baseCur.next()
                moreHead = headCur.next()
                continue
            # mutated
            else:  # (bKey == hKey) and (bVal != hVal)
                mutated.append((hKey, hVal))
                moreBase = baseCur.next()
                moreHead = headCur.next()
                continue

    finally:
        baseCur.close()
        headCur.close()
        TxnRegister().abort_reader_txn(base_env)
        TxnRegister().abort_reader_txn(head_env)

    return DiffOutDB(set(added), set(deleted), set(mutated))


def _raw_from_db_change(changes: Set[Tuple[bytes, bytes]]) -> Changes:
    """Perform conversion for records from db -> raw

    Parameters
    ----------
    changes : Set[Tuple[bytes, bytes]]
        iterable of db formatted key/value pairs

    Returns
    -------
    Changes
        human readable formatted dict of key/value pairs.
    """
    columnKeys, metadataKeys, schemaKeyVals = [], [], []
    for k, v in changes:
        if k[:2] == b'f:':
            columnKeys.append(k)
            continue
        elif k[:2] == b'n:':
            columnKeys.append(k)
            continue
        elif k[:2] == b's:':
            schemaKeyVals.append((k, v))
            continue
        else:
            raise RuntimeError(f'Unknown record type prefix encountered: '
                               f'{k[:2]}. full record => k: {k} & v: {v}')

    columndata = map(dynamic_layout_data_record_from_db_key, columnKeys)
    schemas = {
        schema_column_record_from_db_key(k):
            data_record_digest_val_from_db_val(v) for k, v in schemaKeyVals
    }
    return Changes(schema=schemas, samples=tuple(columndata))


def _all_raw_from_db_changes(outDb: DiffAndConflictsDB) -> DiffAndConflicts:
    """Convert db formatted db diff/conflict results to human readable

    Parameters
    ----------
    outDb : DiffAndConflictsDB
        raw formatted structure containg `diff` and `conflict` fields

    Returns
    -------
    DiffAndConflicts
        Human readable struct containing ``diff`` and ``conflict`` fields.
    """
    it = (outDb.diff.added, outDb.diff.deleted, outDb.diff.mutated)
    out = map(_raw_from_db_change, it)  # significant perf improvement for large commits
    outRawDiff = DiffOut(*out)

    t1 = _raw_from_db_change(outDb.conflict.t1)
    t21 = _raw_from_db_change(outDb.conflict.t21)
    t22 = _raw_from_db_change(outDb.conflict.t22)
    t3 = _raw_from_db_change(outDb.conflict.t3)
    outRawConf = Conflicts(t1=t1, t21=t21, t22=t22, t3=t3, conflict=outDb.conflict.conflict)
    res = DiffAndConflicts(diff=outRawDiff, conflict=outRawConf)
    return res

# ------------------------- Commit Differ -------------------------------------


def _symmetric_difference_keys(pair1: Set[Tuple[bytes, bytes]],
                               pair2: Set[Tuple[bytes, bytes]]
                               ) -> List[Tuple[bytes, bytes]]:
    """Find all keys common to both input pairs AND which have different values.

    Essentially a moddified `symmetric_difference` set operation, which keeps
    track of all seen items. Note: This ignores any `count` tracking values in
    the input tuples (ie. lmdb keys ending in ":")

    Parameters
    ----------
    pair1 : Set[Tuple[bytes, bytes]]
        key/value pairs making up the first set
    pair2 : Set[Tuple[bytes, bytes]]
        key/value pairs making up the second set

    Returns
    -------
    List[Tuple[bytes, bytes]]
        keys which appear in both input pair sets but which have different values.
    """
    seen = set()
    conflict = []
    for k, v in pair1.symmetric_difference(pair2):
        if k in seen:
            conflict.append((k, v))
        else:
            seen.add(k)
    return conflict


def find_conflicts(master_diff: DiffOutDB, dev_diff: DiffOutDB) -> Conflicts:
    """Determine if/which type of conflicting changes occur in diverged commits.

    This function expects the output of :func:`diff_envs` for two commits
    between a base commit.

    Parameters
    ----------
    master_diff : DiffOutDB
        changes (adds, dels, mutations) between base and master HEAD
    dev_diff : DiffOutDB
        changes (adds, dels, mutations) between base and dev HEAD

    Returns
    -------
    Conflicts
        Tuple containing fields for `t1`, `t21`, `t22`, `t3`, and (bool)
        `conflicts` recording output info for if and what type of conflict has
        occured
    """
    t1 = _symmetric_difference_keys(master_diff.added, dev_diff.added)
    t21 = _symmetric_difference_keys(master_diff.deleted, dev_diff.mutated)
    t22 = _symmetric_difference_keys(master_diff.mutated, dev_diff.deleted)
    t3 = _symmetric_difference_keys(master_diff.mutated, dev_diff.mutated)
    isConflict = bool(any([t1, t21, t22, t3]))

    res = Conflicts(t1=t1, t21=t21, t22=t22, t3=t3, conflict=isConflict)
    return res


# ---------------------------- Differ Base  -----------------------------------


class BaseUserDiff(object):

    def __init__(self, branchenv: lmdb.Environment, refenv: lmdb.Environment, *args, **kwargs):

        self._branchenv: lmdb.Environment = branchenv
        self._refenv: lmdb.Environment = refenv

    def _determine_ancestors(self, mHEAD: str, dHEAD: str) -> HistoryDiffStruct:
        """Search the commit history to determine the closest common ancestor.

        The closest common ancestor is important because it serves as the "merge
        base" in a 3-way merge strategy. This is a very naive implementation, but it
        works well enough right now for simple branch histories.

        Parameters
        ----------
        mHEAD : str
            full commit hash to use as the `master` branch head commit
        dHEAD : str
            full commit hash to use as the `dev` branch head commit

        Returns
        -------
        HistoryDiffStruct
            indicating the masterHEAD, devHEAD, ancestorHEAD, and canFF which
            tells if this is a fast-forward-able commit.
        """
        mAncestors = get_commit_ancestors_graph(self._refenv, mHEAD)
        dAncestors = get_commit_ancestors_graph(self._refenv, dHEAD)
        cAncestors = set(mAncestors.keys()).intersection(set(dAncestors.keys()))
        canFF = True if mHEAD in cAncestors else False

        ancestorOrder = []
        for ancestor in cAncestors:
            timeOfCommit = get_commit_spec(self._refenv, ancestor).commit_time
            ancestorOrder.append((ancestor, timeOfCommit))

        ancestorOrder.sort(key=lambda t: t[1], reverse=True)
        commonAncestor = ancestorOrder[0][0]
        res = HistoryDiffStruct(
            masterHEAD=mHEAD, devHEAD=dHEAD, ancestorHEAD=commonAncestor, canFF=canFF)
        return res

    @staticmethod
    def _diff3(a_env: lmdb.Environment,
               m_env: lmdb.Environment,
               d_env: lmdb.Environment) -> DiffAndConflictsDB:
        """Three way diff and conflict finder from ancestor, master, and dev commits.

        Parameters
        ----------
        a_env : lmdb.Environment
            unpacked lmdb environment for the ancestor commit
        m_env : lmdb.Environment
            unpacked lmdb environment for the master commit, current HEAD
        d_env : lmdb.Environment
            unpacked lmdb environment for the dev commit, compare to HEAD

        Returns
        -------
        DiffAndConflictsDB
            structure containing (`additions`, `deletions`, `mutations`) for
            diff, as well as the ConflictRecord struct.
        """
        it = ((a_env, m_env), (a_env, d_env), (d_env, m_env))
        diffs = tuple(starmap(diff_envs, it))  # significant perf improvement by map.
        conflict = find_conflicts(diffs[0], diffs[1])
        return DiffAndConflictsDB(diff=diffs[2], conflict=conflict)

    @staticmethod
    def _diff(a_env: lmdb.Environment, m_env: lmdb.Environment) -> DiffAndConflictsDB:
        """Fast Forward differ from ancestor to master commit.

        Note: this method returns the same MasterDevDiff struct as the three
        way commit diff method, but the `dev` and `conflicts` fields will be
        empty

        Parameters
        ----------
        a_env : lmdb.Environment
            unpacked lmdb environment for the ancestor commit
        m_env : lmdb.Environment
            unpacked lmdb environment for the master commit

        Returns
        -------
        DiffAndConflictsDB
            structure containing (`additions`, `deletions`, `mutations`) for
            the ancestor -> master (head) env diff
        """
        m_diff = diff_envs(a_env, m_env)
        conflict = Conflicts(t1=[], t21=[], t22=[], t3=[], conflict=False)
        return DiffAndConflictsDB(diff=m_diff, conflict=conflict)


# ------------------------ Read-Only Checkouts Only ---------------------------


class ReaderUserDiff(BaseUserDiff):
    """Methods diffing contents of a :class:`~hangar.checkout.ReaderCheckout` instance.

    These provide diffing implementations to compare the current checkout
    ``HEAD`` of a to a branch or commit. The results are generally returned as
    a nested set of named tuples.

    When diffing of commits or branches is performed, if there is not a linear
    history of commits between current ``HEAD`` and the diff commit (ie. a
    history which would permit a ``"fast-forward" merge``), the result field
    named ``conflict`` will contain information on any merge conflicts that
    would exist if staging area ``HEAD`` and the (compared) ``"dev" HEAD`` were
    merged "right now". Though this field is present for all diff comparisons,
    it can only contain non-empty values in the cases where a three way merge
    would need to be performed.

    ::

       Fast Forward is Possible
       ========================

           (master)          (foo)
       a ----- b ----- c ----- d


       3-Way Merge Required
       ====================

                            (master)
       a ----- b ----- c ----- d
               \\
                \\               (foo)
                 \\----- ee ----- ff
    """

    def __init__(self, commit_hash, *args, **kwargs):

        super().__init__(*args, **kwargs)
        self._commit_hash = commit_hash

    def _run_diff(self, dev_commit_hash: str) -> DiffAndConflictsDB:
        """Compute diff between head and commit hash, returning DB formatted results

        Parameters
        ----------
        dev_commit_hash : str
            hash of the commit to be used as the comparison.

        Returns
        -------
        DiffAndConflictsDB
            two-tuple of `diff`, `conflict` (if any) calculated in the diff
            algorithm.
        """
        hist = self._determine_ancestors(self._commit_hash, dev_commit_hash)
        mH, dH, aH = hist.masterHEAD, hist.devHEAD, hist.ancestorHEAD
        with tmp_cmt_env(self._refenv, mH) as m_env, tmp_cmt_env(self._refenv, dH) as d_env:
            if hist.canFF is True:
                outDb = self._diff(m_env, d_env)
            else:
                with tmp_cmt_env(self._refenv, aH) as a_env:
                    outDb = self._diff3(a_env, m_env, d_env)
        return outDb

    def commit(self, dev_commit_hash: str) -> DiffAndConflicts:
        """Compute diff between HEAD and commit hash, returning user-facing results.

        Parameters
        ----------
        dev_commit_hash : str
            hash of the commit to be used as the comparison.

        Returns
        -------
        DiffAndConflicts
            two-tuple of ``diff``, ``conflict`` (if any) calculated in the diff
            algorithm.

        Raises
        ------
        ValueError
            if the specified ``dev_commit_hash`` is not a valid commit reference.
        """
        if not check_commit_hash_in_history(self._refenv, dev_commit_hash):
            msg = f'HANGAR VALUE ERROR: dev_commit_hash: {dev_commit_hash} does not exist'
            raise ValueError(msg)

        outDb = self._run_diff(dev_commit_hash=dev_commit_hash)
        outRaw = _all_raw_from_db_changes(outDb)
        return outRaw

    def branch(self, dev_branch: str) -> DiffAndConflicts:
        """Compute diff between HEAD and branch name, returning user-facing results.

        Parameters
        ----------
        dev_branch : str
            name of the branch whose HEAD will be used to calculate the diff of.

        Returns
        -------
        DiffAndConflicts
            two-tuple of ``diff``, ``conflict`` (if any) calculated in the diff
            algorithm.

        Raises
        ------
        ValueError
            If the specified `dev_branch` does not exist.
        """
        branchNames = get_branch_names(self._branchenv)
        if dev_branch in branchNames:
            dHEAD = get_branch_head_commit(self._branchenv, dev_branch)
        else:
            msg = f'HANGAR VALUE ERROR: dev_branch: {dev_branch} invalid branch name'
            raise ValueError(msg)

        outDb = self._run_diff(dev_commit_hash=dHEAD)
        outRaw = _all_raw_from_db_changes(outDb)
        return outRaw


# ---------------------- Write Enabled Checkouts Only -------------------------


class WriterUserDiff(BaseUserDiff):
    """Methods diffing contents of a :class:`~hangar.checkout.WriterCheckout` instance.

    These provide diffing implementations to compare the current ``HEAD`` of a
    checkout to a branch, commit, or the staging area ``"base"`` contents. The
    results are generally returned as a nested set of named tuples. In
    addition, the :meth:`status` method is implemented which can be used to
    quickly determine if there are any uncommitted changes written in the
    checkout.

    When diffing of commits or branches is performed, if there is not a linear
    history of commits between current ``HEAD`` and the diff commit (ie. a
    history which would permit a ``"fast-forward" merge``), the result field
    named ``conflict`` will contain information on any merge conflicts that
    would exist if staging area ``HEAD`` and the (compared) ``"dev" HEAD`` were
    merged "right now". Though this field is present for all diff comparisons,
    it can only contain non-empty values in the cases where a three way merge
    would need to be performed.

    ::

       Fast Forward is Possible
       ========================

           (master)          (foo)
       a ----- b ----- c ----- d


       3-Way Merge Required
       ====================

                            (master)
       a ----- b ----- c ----- d
               \\
                \\               (foo)
                 \\----- ee ----- ff
    """

    def __init__(self, stageenv: lmdb.Environment, branch_name: str, *args, **kwargs):

        super().__init__(*args, **kwargs)
        self._stageenv: lmdb.Environment = stageenv
        self._branch_name: str = branch_name

    def _run_diff(self, dev_commit_hash: str) -> DiffAndConflictsDB:
        """Compute diff between head and commit, returning DB formatted results.

        Parameters
        ----------
        dev_commit_hash : str
            hash of the commit to be used as the comparison.

        Returns
        -------
        DiffAndConflictsDB
            two-tuple of `diff`, `conflict` (if any) calculated in the diff
            algorithm.
        """
        commit_hash = get_branch_head_commit(self._branchenv, self._branch_name)
        hist = self._determine_ancestors(commit_hash, dev_commit_hash)
        with tmp_cmt_env(self._refenv, hist.devHEAD) as d_env:
            if hist.canFF is True:
                res = self._diff(self._stageenv, d_env)
            else:
                with tmp_cmt_env(self._refenv, hist.ancestorHEAD) as a_env:
                    res = self._diff3(a_env, self._stageenv, d_env)
        return res

    def commit(self, dev_commit_hash: str) -> DiffAndConflicts:
        """Compute diff between HEAD and commit, returning user-facing results.

        Parameters
        ----------
        dev_commit_hash : str
            hash of the commit to be used as the comparison.

        Returns
        -------
        DiffAndConflicts
            two-tuple of ``diff``, ``conflict`` (if any) calculated in the diff
            algorithm.

        Raises
        ------
        ValueError
            if the specified ``dev_commit_hash`` is not a valid commit reference.
        """
        if not check_commit_hash_in_history(self._refenv, dev_commit_hash):
            msg = f'HANGAR VALUE ERROR: dev_commit_hash: {dev_commit_hash} does not exist'
            raise ValueError(msg)

        outDb = self._run_diff(dev_commit_hash=dev_commit_hash)
        outRaw = _all_raw_from_db_changes(outDb)
        return outRaw

    def branch(self, dev_branch: str) -> DiffAndConflicts:
        """Compute diff between HEAD and branch, returning user-facing results.

        Parameters
        ----------
        dev_branch : str
            name of the branch whose HEAD will be used to calculate the diff of.

        Returns
        -------
        DiffAndConflicts
            two-tuple of ``diff``, ``conflict`` (if any) calculated in the diff
            algorithm.

        Raises
        ------
        ValueError
            If the specified ``dev_branch`` does not exist.
        """
        branchNames = get_branch_names(self._branchenv)
        if dev_branch in branchNames:
            dHEAD = get_branch_head_commit(self._branchenv, dev_branch)
        else:
            msg = f'HANGAR VALUE ERROR: dev_branch: {dev_branch} invalid branch name'
            raise ValueError(msg)

        outDb = self._run_diff(dev_commit_hash=dHEAD)
        outRaw = _all_raw_from_db_changes(outDb)
        return outRaw

    def staged(self) -> DiffAndConflicts:
        """Return diff of staging area to base, returning user-facing results.

        Returns
        -------
        DiffAndConflicts
            two-tuple of ``diff``, ``conflict`` (if any) calculated in the diff
            algorithm.
        """
        commit_hash = get_branch_head_commit(self._branchenv, self._branch_name)
        with tmp_cmt_env(self._refenv, commit_hash) as base_env:
            outDb = self._diff(base_env, self._stageenv)
        outRaw = _all_raw_from_db_changes(outDb)
        return outRaw

    def status(self) -> str:
        """Determine if changes have been made in the staging area

        If the contents of the staging area and it's parent commit are the
        same, the status is said to be "CLEAN". If even one column or
        metadata record has changed however, the status is "DIRTY".

        Returns
        -------
        str
            "CLEAN" if no changes have been made, otherwise "DIRTY"
        """
        head_commit = get_branch_head_commit(self._branchenv, self._branch_name)
        if head_commit == '':
            base_refs = ()
        else:
            base_refs = get_commit_ref(self._refenv, head_commit)

        stage_refs = tuple(RecordQuery(self._stageenv)._traverse_all_records())
        status = 'DIRTY' if (base_refs != stage_refs) else 'CLEAN'
        return status



# header file for record containers

cdef class CompatibleData:
    cdef readonly bint compatible
    cdef readonly str reason


cdef class ColumnSchemaKey:
    cdef readonly str column
    cdef readonly str layout


cdef class FlatColumnDataKey:
    cdef readonly str column
    cdef str _sample
    cdef bint _s_int


cdef class NestedColumnDataKey:
    cdef readonly str column
    cdef str _sample, _subsample
    cdef bint _s_int, _ss_int


cdef class DataRecordVal:
    cdef readonly str digest



from collections import defaultdict
from functools import partial
import secrets
import string

import pytest
import numpy as np


def assert_equal(expected, actual):
    if isinstance(expected, str):
        assert expected == actual
    elif isinstance(expected, np.ndarray):
        assert np.allclose(expected, actual)
        assert expected.dtype == actual.dtype
    else:
        raise TypeError(f'unknown type of data {type(expected)}')


def ndarray_generate_data_fixed_shape(shape, dtype, low=0, high=255):
    arr = np.random.randint(low, high, size=shape, dtype=dtype)
    return arr


def ndarray_generate_data_variable_shape(shape, dtype, low=0, high=255):
    arr_dims = []
    for dim in shape:
        valid_dim_shapes = [i for i in range(1, dim + 1)]
        dimsize = secrets.choice(valid_dim_shapes)
        arr_dims.append(dimsize)
    arr_dims = tuple(arr_dims)
    arr = np.random.randint(low, high, size=arr_dims, dtype=dtype)
    return arr


def str_generate_data_variable_shape(
        length=20, *, _ALPHABET=''.join([string.ascii_letters, string.digits, string.punctuation, ' '])
):
    tokens = [secrets.choice(_ALPHABET) for i in range(length)]
    res = ''.join(tokens)
    return res


column_settings = {
    'ndarray': {
        'fixed_shape': ['00', '01', '10'],
        'variable_shape': ['00', '10'],
    },
    'str': {
        'variable_shape': ['30']
    }
}


column_data_generators = {
    'ndarray': {
        'fixed_shape': ndarray_generate_data_fixed_shape,
        'variable_shape': ndarray_generate_data_variable_shape,
    },
    'str': {
        'variable_shape': str_generate_data_variable_shape
    }
}


column_layouts = {
    'ndarray': ['flat', 'nested'],
    'str': ['flat', 'nested']
}


def add_data_to_column(col, data_gen, nsamples, nsubsamples=None):
    column_data = {}
    for samp in range(nsamples):
        if nsubsamples is None:
            data = data_gen()
            column_data[samp] = data
            col[samp] = data
        else:
            column_data[samp] = {}
            for subsamp in range(nsubsamples):
                data = data_gen()
                column_data[samp][subsamp] = data
            col[samp] = column_data[samp]
    return column_data


@pytest.fixture(params=[1, 3])
def num_samples_gen(request):
    return request.param


@pytest.fixture(params=[1, 3])
def num_subsamples_gen(request):
    return request.param


@pytest.fixture()
def column_permutation_repo(repo, num_samples_gen, num_subsamples_gen):
    co = repo.checkout(write=True)
    nsamp = num_samples_gen
    nsubs = num_subsamples_gen

    column_name_partials = {}
    column_data_copy = defaultdict(dict)
    shape = (4, 4)
    dtype = np.uint8
    for col_dtype, schema_settings in column_settings.items():
        for layout in column_layouts[col_dtype]:
            for schema_type, valid_backends in schema_settings.items():
                for backend in valid_backends:
                    name = f'{col_dtype}_{layout}_{schema_type}_{backend}'
                    generator = column_data_generators[col_dtype][schema_type]
                    has_subs = False if layout == 'flat' else True
                    is_var = False if schema_type == 'fixed_shape' else True

                    if col_dtype == 'ndarray':
                        col = co.add_ndarray_column(name,
                                                       shape=shape,
                                                       dtype=dtype,
                                                       variable_shape=is_var,
                                                       contains_subsamples=has_subs,
                                                       backend=backend)
                        data_partial = partial(generator, shape, dtype)
                        if layout == 'flat':
                            column_data_copy[name] = add_data_to_column(col, data_partial, nsamp)
                        elif layout == 'nested':
                            column_data_copy[name] = add_data_to_column(col, data_partial, nsamp, nsubs)
                        else:
                            raise ValueError(f'invalid layout {layout}')
                    elif col_dtype == 'str':
                        col = co.add_str_column(name, contains_subsamples=has_subs, backend=backend)
                        data_partial = partial(generator)
                        if layout == 'flat':
                            column_data_copy[name] = add_data_to_column(col, data_partial, nsamp)
                        elif layout == 'nested':
                            column_data_copy[name] = add_data_to_column(col, data_partial, nsamp, nsubs)
                    else:
                        raise ValueError(f'column dtype {col_dtype} invalid')

                    column_name_partials[name] = data_partial

    co.commit('first')
    co.close()
    yield repo, column_data_copy, column_name_partials


@pytest.fixture(params=[True, False])
def column_permutations_read_write_checkout(request, column_permutation_repo):
    repo, column_data, column_data_partials = column_permutation_repo
    co = repo.checkout(write=request.param)
    yield co, column_data, column_data_partials
    co.close()


@pytest.fixture()
def column_permutations_write_checkout(column_permutation_repo):
    repo, column_data, column_data_partials = column_permutation_repo
    co = repo.checkout(write=True)
    yield co, column_data, column_data_partials
    co.close()


def test_read_data_from_column_permutations(column_permutations_read_write_checkout):
    co, column_data, column_data_partials = column_permutations_read_write_checkout

    assert len(co.columns) == len(column_data)
    for column_name, column_samples in column_data.items():
        assert column_name in co.columns
        col = co[column_name]
        assert len(column_samples) == len(col)

        for sample_key, sample_value in column_samples.items():
            if not isinstance(sample_value, dict):
                recorded = col[sample_key]
                assert_equal(sample_value, recorded)
            else:
                col_samp = col[sample_key]
                assert len(sample_value) == len(col_samp)
                for subsample_key, subsample_value in sample_value.items():
                    recorded = col_samp[subsample_key]
                    assert_equal(subsample_value, recorded)


def test_write_data_to_column_permutations(
        column_permutations_write_checkout, num_samples_gen, num_subsamples_gen
):
    co, column_data, column_data_partials = column_permutations_write_checkout

    for column_name in co.columns:
        col = co[column_name]
        data_gen = column_data_partials[column_name]
        if col.column_layout == 'flat':
            for samp in range(num_samples_gen):
                data = data_gen()
                column_data[column_name][str(samp)] = data
                col[str(samp)] = data
        elif col.column_layout == 'nested':
            for samp in range(num_samples_gen):
                column_data[column_name][str(samp)] = {}
                for ssamp in range(num_subsamples_gen):
                    data = data_gen()
                    column_data[column_name][str(samp)][str(ssamp)] = data
                col[str(samp)] = column_data[column_name][str(samp)]
        else:
            raise ValueError(f'unknown layout option {col.column_layout}')

    assert len(co.columns) == len(column_data)
    for column_name, column_samples in column_data.items():
        assert column_name in co.columns
        col = co[column_name]
        assert len(column_samples) == len(col)

        for sample_key, sample_value in column_samples.items():
            if not isinstance(sample_value, dict):
                recorded = col[sample_key]
                assert_equal(sample_value, recorded)
            else:
                col_samp = col[sample_key]
                assert len(sample_value) == len(col_samp)
                for subsample_key, subsample_value in sample_value.items():
                    recorded = col_samp[subsample_key]
                    assert_equal(subsample_value, recorded)


def test_merge_write_data_to_column_permutations(
        column_permutation_repo, num_samples_gen, num_subsamples_gen
):
    repo, column_data, column_data_partials = column_permutation_repo
    repo.create_branch('testbranch')

    # Write new data to master branch
    co = repo.checkout(write=True, branch='master')
    for column_name in co.columns:
        col = co[column_name]
        data_gen = column_data_partials[column_name]
        if col.column_layout == 'flat':
            for samp in range(num_samples_gen):
                data = data_gen()
                column_data[column_name][str(samp)] = data
                col[str(samp)] = data
        elif col.column_layout == 'nested':
            for samp in range(num_samples_gen):
                column_data[column_name][str(samp)] = {}
                for ssamp in range(num_subsamples_gen):
                    data = data_gen()
                    column_data[column_name][str(samp)][str(ssamp)] = data
                col[str(samp)] = column_data[column_name][str(samp)]
        else:
            raise ValueError(f'unknown layout option {col.column_layout}')
    co.commit('commit on master adding data')
    co.close()

    # Write new data to testbranch branch
    co = repo.checkout(write=True, branch='testbranch')
    for column_name in co.columns:
        col = co[column_name]
        data_gen = column_data_partials[column_name]
        if col.column_layout == 'flat':
            for samp in range(num_samples_gen):
                data = data_gen()
                column_data[column_name][f'_{samp}'] = data
                col[f'_{samp}'] = data
        elif col.column_layout == 'nested':
            for samp in range(num_samples_gen):
                column_data[column_name][f'_{samp}'] = {}
                for ssamp in range(num_subsamples_gen):
                    data = data_gen()
                    column_data[column_name][f'_{samp}'][f'_{ssamp}'] = data
                col[f'_{samp}'] = column_data[column_name][f'_{samp}']
        else:
            raise ValueError(f'unknown layout option {col.column_layout}')
    co.commit('commit on master adding data')
    co.close()

    # Merge and check that union of all data added is present
    repo.merge('merge commit', 'master', 'testbranch')

    co = repo.checkout(write=True, branch='master')
    assert len(co.columns) == len(column_data)
    for column_name, column_samples in column_data.items():
        assert column_name in co.columns
        col = co[column_name]
        assert len(column_samples) == len(col)

        for sample_key, sample_value in column_samples.items():
            if not isinstance(sample_value, dict):
                recorded = col[sample_key]
                assert_equal(sample_value, recorded)
            else:
                col_samp = col[sample_key]
                assert len(sample_value) == len(col_samp)
                for subsample_key, subsample_value in sample_value.items():
                    recorded = col_samp[subsample_key]
                    assert_equal(subsample_value, recorded)
    co.close()




"""
Portions of this code have been taken and modified from the book:

Beazley, D. and B. K. Jones (2013). Python Cookbook, O’Reilly Media, Inc.

Chapter: 8.13. Implementing a Data Model or Type System

===============================================================================

Problem
-------

You want to define various kinds of data structures, but want to enforce
constraints on the values that are allowed to be assigned to certain
attributes.

Solution
--------

In this problem, you are basically faced with the task of placing checks or
assertions on the setting of certain instance attributes. To do this, you need
to customize the setting of attributes on a per-attribute basis. To do this,
you should use descriptors.

This recipe involves a number of advanced techniques, including descriptors,
mixin classes, the use of super(), class decorators, and metaclasses. Covering
the basics of all those topics is beyond what can be covered here; However,
there are a number of subtle points worth noting.

First, in the Descriptor base class, you will notice that there is a __set__()
method, but no corresponding __get__(). If a descriptor will do nothing more
than extract an identically named value from the underlying instance
dictionary, defining __get__() is unnecessary. In fact, defining __get__() will
just make it run slower. Thus, this recipe only focuses on the implementation
of __set__().

The overall design of the various descriptor classes is based on mixin classes.
For example, the Unsigned and MaxSized classes are meant to be mixed with the
other descriptor classes derived from Typed. To handle a specific kind of data
type, multiple inheritance is used to combine the desired functionality.

You will also notice that all __init__() methods of the various descriptors
have been programmed to have an identical signature involving keyword arguments
**opts. The class for MaxSized looks for its required attribute in opts, but
simply passes it along to the Descriptor base class, which actually sets it.
One tricky part about composing classes like this (especially mixins), is that
you don’t always know how the classes are going to be chained together or what
super() will invoke. For this reason, you need to make it work with any
possible combination of classes.

The definitions of the various type classes such as Integer, Float, and String
illustrate a useful technique of using class variables to customize an
implementation. The Ty ped descriptor merely looks for an expected_type
attribute that is provided by each of those subclasses.

The use of a class decorator or metaclass is often useful for simplifying the
specification by the user.

The code for the class decorator and metaclass simply scan the class dictionary
looking for descriptors. When found, they simply fill in the descriptor name
based on the key value.

As a final twist, a class decorator approach can also be used as a replacement
for mixin classes, multiple inheritance, and tricky use of the super() function

The classes defined in this alternative formulation work in exactly the same
manner as before (none of the earlier example code changes) except that
everything runs much faster. For example, a simple timing test of setting a
typed attribute reveals that the class decorator approach runs almost 100%
faster than the approach using mixins.
"""
from typing import Sequence


class Descriptor:
    # Base class. Uses a descriptor to set a value
    def __init__(self, name=None, **opts):
        self.name = name
        self.__dict__.update(opts)

    def __set__(self, instance, value):
        instance.__dict__[self.name] = value


def Typed(expected_type, cls=None):
    # Decorator for applying type checking
    if cls is None:
        return lambda cls: Typed(expected_type, cls)

    super_set = cls.__set__

    def __set__(self, instance, value):
        if not isinstance(value, expected_type):
            raise TypeError('expected ' + str(expected_type))
        super_set(self, instance, value)

    cls.__set__ = __set__
    return cls


def TypedSequence(expected_element_types, cls=None):
    # Decorator enforcing that all elements in an sequence are specific type(s).
    # using the python ABC definition of "Sequence" (list, tuple)
    # https://docs.python.org/3/library/collections.abc.html#collections.abc.Sequence
    if cls is None:
        return lambda cls: TypedSequence(expected_element_types, cls)

    super_set = cls.__set__
    def __set__(self, instance, value):
        if not isinstance(value, Sequence):
            raise TypeError(f'input is not Sequence type, recieved {type(value)}')
        elif not all([isinstance(el, expected_element_types) for el in value]):
            raise TypeError(f'not all elements are {expected_element_types} type(s) in {value}')
        super_set(self, instance, value)
    cls.__set__ = __set__
    return cls


def OneOf(expected_values, cls=None):
    # Decorator for enforcing values
    if cls is None:
        return lambda cls: OneOf(expected_values, cls)

    super_set = cls.__set__
    def __set__(self, instance, value):
        if value not in expected_values:
            raise ValueError(f'expected one of {expected_values} recieved {value}')
        super_set(self, instance, value)
    cls.__set__ = __set__
    return cls


def MaxSized(cls):
    # Decorator for allowing sized values
    super_init = cls.__init__
    def __init__(self, name=None, **opts):
        if 'size' not in opts:
            raise TypeError('missing size option')
        self.size = opts['size']
        super_init(self, name, **opts)
    cls.__init__ = __init__
    super_set = cls.__set__
    def __set__(self, instance, value):
        if len(value) > self.size:
            raise ValueError('size must be < ' + str(self.size))
        super_set(self, instance, value)
    cls.__set__ = __set__
    return cls


def DictItems(expected_keys_required, expected_values, cls=None):
    # check a dictionary for the existence of keys. expected_keys should be a dictionary of keys,
    # with bool values set to indicate if they are required or not. expected_values should be
    # mapping of same keys to list of acceptable values.
    if cls is None:
        return lambda cls: DictItems(expected_keys_required, expected_values, cls)

    super_set = cls.__set__
    def __set__(self, instance, value):
        if not isinstance(value, dict):
            raise TypeError(f'expected {dict}, recieved {type(value)}')
        for expected_key, required in expected_keys_required.items():
            try:
                if value[expected_key] not in expected_values[expected_key]:
                    raise ValueError(f'{value[expected_key]} invalid for key {expected_key}')
            except KeyError as e:
                if required:
                    raise e
        for recieved_key in value.keys():
            if recieved_key not in expected_keys_required:
                raise TypeError(f'Not supposed to have key {recieved_key}')
        super_set(self, instance, value)
    cls.__set__ = __set__
    return cls


@Typed(str)
class String(Descriptor):
    pass


@DictItems(expected_keys_required={},
           expected_values={},)
class EmptyDict(Descriptor):
    pass


@Typed((dict, type(None)))
class OptionalDict(Descriptor):
    pass


@Typed((str, type(None)))
class OptionalString(Descriptor):
    pass


@Typed(tuple)
class Tuple(Descriptor):
    pass


@MaxSized
@TypedSequence(int)
class SizedIntegerTuple(Tuple):
    pass


class checkedmeta(type):
    # A metaclass that applies checking
    def __new__(cls, clsname, bases, methods):
        # Attach attribute names to the descriptors
        for key, value in methods.items():
            if isinstance(value, Descriptor):
                value.name = key
        return type.__new__(cls, clsname, bases, methods)



from os.path import join as pjoin
from os import mkdir
import pytest
import sys
import numpy as np
from hangar import Repository


try:
    import torch
    from torch.utils.data import DataLoader
    from hangar import make_torch_dataset
    torchExists = True
except (ImportError, ModuleNotFoundError):
    torchExists = False


@pytest.mark.skipif(torchExists,
                    reason='pytorch is installed in the test environment.')
def test_no_torch_installed_raises_error_on_dataloader_import():
    with pytest.raises(ImportError):
        from hangar import make_torch_dataset
        make_torch_dataset(None)


@pytest.mark.skipif(not torchExists,
                    reason='pytorch is not installed in the test environment.')
class TestTorchDataLoader(object):

    def test_warns_experimental(self, repo_20_filled_samples):
        repo = repo_20_filled_samples
        co = repo.checkout()
        first_aset = co.columns['writtenaset']
        second_aset = co.columns['second_aset']
        with pytest.warns(UserWarning, match='Dataloaders are experimental'):
            make_torch_dataset([first_aset, second_aset])
        co.close()

    @pytest.mark.filterwarnings("ignore:Dataloaders are experimental")
    def test_warns_arrayset_sample_size_mismatch(self, repo_20_filled_samples):
        repo = repo_20_filled_samples
        co = repo.checkout(write=True)
        second_aset = co.columns['second_aset']
        del second_aset['10']
        co.commit('deleting')
        co.close()

        co = repo.checkout()
        first_aset = co.columns['writtenaset']
        second_aset = co.columns['second_aset']
        with pytest.warns(UserWarning, match='Columns do not contain equal number of samples'):
            make_torch_dataset([first_aset, second_aset])
        co.close()

    @pytest.mark.filterwarnings("ignore:Dataloaders are experimental")
    def test_multiple_dataset_loader(self, repo_20_filled_samples):
        repo = repo_20_filled_samples
        co = repo.checkout(write=True)
        second_aset = co.columns['second_aset']
        del second_aset['10']
        co.commit('deleting')
        co.close()

        co = repo.checkout()
        first_aset = co.columns['writtenaset']
        second_aset = co.columns['second_aset']
        with pytest.raises(ValueError):
            # emtpy list
            make_torch_dataset([])
        with pytest.raises(TypeError):
            # if more than one dataset, those should be in a list/tuple
            make_torch_dataset(first_aset, first_aset)

        with pytest.warns(UserWarning, match='Columns do not contain equal number of samples'):
            torch_dset = make_torch_dataset([first_aset, second_aset])
        loader = DataLoader(torch_dset, batch_size=6, drop_last=True)
        total_samples = 0
        for dset1, dset2 in loader:
            total_samples += dset1.shape[0]
            assert dset1.shape == (6, 5, 7)
            assert dset2.shape == (6, 5, 7)
        assert total_samples == 18  # drop last is True
        co.close()

    @pytest.mark.filterwarnings("ignore:Dataloaders are experimental")
    def test_dataset_loader_fails_with_write_enabled_checkout(self, repo_20_filled_samples):
        repo = repo_20_filled_samples
        co = repo.checkout(write=True)
        first_aset = co.columns['writtenaset']
        second_aset = co.columns['second_aset']
        with pytest.raises(TypeError):
            make_torch_dataset([first_aset, second_aset])
        co.close()

    @pytest.mark.filterwarnings("ignore:Dataloaders are experimental")
    def test_with_keys(self, repo_20_filled_samples):
        repo = repo_20_filled_samples
        co = repo.checkout()
        aset = co.columns['writtenaset']

        # with keys
        keys = ['2', '4', '5', '6', '7', '9', '15', '18', '19']
        bad_tensor0 = aset['0']
        bad_tensor1 = aset['1']
        bad_tensor3 = aset['3']
        bad_tensor8 = aset['8']

        torch_dset = make_torch_dataset(aset, keys=keys)
        loader = DataLoader(torch_dset, batch_size=3)
        total_batches = 0
        for batch in loader:
            assert batch[0].size(0) == 3
            total_batches += 1
            for sample in batch:
                assert not np.allclose(sample, bad_tensor0)
                assert not np.allclose(sample, bad_tensor1)
                assert not np.allclose(sample, bad_tensor3)
                assert not np.allclose(sample, bad_tensor8)
        assert total_batches == 3
        co.close()

    @pytest.mark.filterwarnings("ignore:Dataloaders are experimental")
    def test_with_index_range(self, repo_20_filled_samples):
        repo = repo_20_filled_samples
        co = repo.checkout()
        aset = co.columns['writtenaset']

        # with keys
        bad_tensor0 = aset['0']
        bad_tensor1 = aset['1']

        # with index range
        index_range = slice(2, 20)
        torch_dset = make_torch_dataset(aset, index_range=index_range)
        loader = DataLoader(torch_dset, batch_size=3)
        total_batches = 0
        for batch in loader:
            assert batch[0].size(0) == 3
            total_batches += 1
            for sample in batch:
                assert not np.allclose(sample, bad_tensor0)
                assert not np.allclose(sample, bad_tensor1)
        assert total_batches == 6
        co.close()

    @pytest.mark.filterwarnings("ignore:Dataloaders are experimental")
    def test_field_names(self, repo_20_filled_samples):
        repo = repo_20_filled_samples
        co = repo.checkout()
        first_aset = co.columns['writtenaset']
        second_aset = co.columns['second_aset']
        with pytest.raises(ValueError):  # number of dsets and field_names are different
            make_torch_dataset([first_aset, second_aset], field_names=('input',))
        with pytest.raises(TypeError):  # field_names's type is wrong
            make_torch_dataset([first_aset, second_aset], field_names={'input': '', 'target': ''})
        torch_dset = make_torch_dataset([first_aset, second_aset], field_names=('input', 'target'))
        assert len(torch_dset) == 20
        loader = DataLoader(torch_dset, batch_size=5)
        for sample in loader:
            assert type(sample).__name__ == 'BatchTuple_input_target'
            assert sample._fields == ('input', 'target')
        co.close()

    @pytest.mark.filterwarnings("ignore:Dataloaders are experimental")
    def test_lots_of_data_with_multiple_backend(self, repo_300_filled_samples):
        repo = repo_300_filled_samples
        co = repo.checkout()
        aset = co.columns['aset']
        torch_dset = make_torch_dataset([aset])
        loader = DataLoader(torch_dset, batch_size=10, drop_last=True)
        for data in loader:
            assert type(data).__name__ == 'BatchTuple_aset'
            assert data.aset.shape == (10, 5, 7)
        co.close()

    @pytest.mark.xfail(sys.platform == "win32",
                       strict=True,
                       reason="multiprocess workers does not run on windows")
    @pytest.mark.filterwarnings("ignore:Dataloaders are experimental")
    def test_lots_of_data_with_multiple_backend_multiple_worker_dataloader(self,
                                                                           repo_300_filled_samples):
        repo = repo_300_filled_samples
        co = repo.checkout()
        aset = co.columns['aset']
        torch_dset = make_torch_dataset([aset])
        loader = DataLoader(torch_dset, batch_size=10, drop_last=True, num_workers=2)
        for data in loader:
            assert type(data).__name__ == 'BatchTuple_aset'
            assert data.aset.shape == (10, 5, 7)
        co.close()

    @pytest.mark.xfail(sys.platform == "win32",
                       strict=True,
                       reason="multiprocess workers does not run on windows")
    @pytest.mark.filterwarnings("ignore:Dataloaders are experimental")
    def test_two_aset_loader_two_worker_dataloader(self, repo_20_filled_samples):
        repo = repo_20_filled_samples
        co = repo.checkout()
        first_aset = co.columns['writtenaset']
        second_aset = co.columns['second_aset']
        torch_dset = make_torch_dataset([first_aset, second_aset])
        loader = DataLoader(torch_dset, batch_size=2, drop_last=True, num_workers=2)
        count = 0
        for asets_batch in loader:
            assert type(asets_batch).__name__ == 'BatchTuple_writtenaset_second_aset'
            assert isinstance(asets_batch, tuple)
            assert len(asets_batch) == 2
            assert asets_batch._fields == ('writtenaset', 'second_aset')
            assert asets_batch.writtenaset.shape == (2, 5, 7)
            assert asets_batch.second_aset.shape == (2, 5, 7)
            assert np.allclose(asets_batch.writtenaset, -asets_batch.second_aset)
            count += 1
        assert count == 10

    @pytest.mark.filterwarnings("ignore:Dataloaders are experimental")
    @pytest.mark.filterwarnings("ignore:Column.* writtenaset contains `reference-only` samples")
    def test_local_without_data_fails_no_common_no_local(self, written_two_cmt_server_repo, managed_tmpdir):
        new_tmpdir = pjoin(managed_tmpdir, 'new')
        mkdir(new_tmpdir)
        server, _ = written_two_cmt_server_repo
        repo = Repository(path=new_tmpdir, exists=False)
        repo.clone('name', 'a@b.c', server, remove_old=True)
        co = repo.checkout()
        aset = co.columns['writtenaset']
        with pytest.raises(ValueError):
            torch_dset = make_torch_dataset(aset)
        co.close()
        repo._env._close_environments()

    @pytest.mark.filterwarnings("ignore:Dataloaders are experimental")
    @pytest.mark.filterwarnings("ignore:Column.* writtenaset contains `reference-only` samples")
    def test_local_without_data_fails_no_common(self, written_two_cmt_server_repo, managed_tmpdir):
        new_tmpdir = pjoin(managed_tmpdir, 'new')
        mkdir(new_tmpdir)
        server, _ = written_two_cmt_server_repo
        repo = Repository(path=new_tmpdir, exists=False)
        repo.clone('name', 'a@b.c', server, remove_old=True)
        co = repo.checkout()
        aset = co.columns['writtenaset']
        with pytest.raises(KeyError):
            torch_dset = make_torch_dataset(aset, keys=['1', -1])
        co.close()
        repo._env._close_environments()

    @pytest.mark.filterwarnings("ignore:Dataloaders are experimental")
    @pytest.mark.filterwarnings("ignore:Column.* writtenaset contains `reference-only` samples")
    def test_local_without_data_fails_data_unavailable(self, written_two_cmt_server_repo, managed_tmpdir):
        new_tmpdir = pjoin(managed_tmpdir, 'new')
        mkdir(new_tmpdir)
        server, _ = written_two_cmt_server_repo
        repo = Repository(path=new_tmpdir, exists=False)
        repo.clone('name', 'a@b.c', server, remove_old=True)
        co = repo.checkout()
        aset = co.columns['writtenaset']
        with pytest.raises(FileNotFoundError):
            torch_dset = make_torch_dataset(aset, keys=['1', '2'])
        co.close()
        repo._env._close_environments()


try:
    import warnings
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', category=DeprecationWarning)
        import tensorflow as tf
    tf.compat.v1.enable_eager_execution()
    from hangar import make_tf_dataset
    tfExists = True
except (ImportError, ModuleNotFoundError):
    tfExists = False


@pytest.mark.skipif(tfExists,
                    reason='tensorflow is installed in the test environment.')
def test_no_tf_installed_raises_error_on_dataloader_import():
    with pytest.raises(ImportError):
        from hangar import make_tf_dataset
        make_tf_dataset(None)


@pytest.mark.skipif(not tfExists,
                    reason='tensorflow is not installed in the test environment.')
class TestTfDataLoader(object):

    def test_warns_experimental(self, repo_20_filled_samples):
        repo = repo_20_filled_samples
        co = repo.checkout()
        first_aset = co.columns['writtenaset']
        second_aset = co.columns['second_aset']
        with pytest.warns(UserWarning, match='Dataloaders are experimental'):
            make_tf_dataset([first_aset, second_aset])
        co.close()

    @pytest.mark.filterwarnings("ignore:Dataloaders are experimental")
    def test_wans_arrayset_sample_size_mismatch(self, repo_20_filled_samples):
        repo = repo_20_filled_samples
        co = repo.checkout(write=True)
        second_aset = co.columns['second_aset']
        del second_aset['10']
        co.commit('deleting')
        co.close()

        co = repo.checkout()
        first_aset = co.columns['writtenaset']
        second_aset = co.columns['second_aset']
        with pytest.warns(UserWarning, match='Columns do not contain equal number of samples'):
            make_tf_dataset([first_aset, second_aset])
        co.close()

    @pytest.mark.filterwarnings("ignore:Dataloaders are experimental")
    def test_dataset_loader(self, repo_20_filled_samples):
        repo = repo_20_filled_samples
        co = repo.checkout()
        first_aset = co.columns['writtenaset']
        second_aset = co.columns['second_aset']

        # multiple datasets
        tf_dset = make_tf_dataset([first_aset, second_aset])
        tf_dset = tf_dset.batch(6)
        for dset1, dset2 in tf_dset.take(2):
            assert dset1.shape == tf.TensorShape((6, 5, 7))
            assert dset2.shape == tf.TensorShape((6, 5, 7))
        co.close()

    @pytest.mark.filterwarnings("ignore:Dataloaders are experimental")
    def test_with_keys(self, repo_20_filled_samples):
        repo = repo_20_filled_samples
        co = repo.checkout()
        aset = co.columns['writtenaset']

        # with keys
        keys = ['2', '4', '5', '6', '7', '9', '15', '18', '19']
        bad_tensor0 = aset['0']
        bad_tensor1 = aset['1']
        bad_tensor3 = aset['3']
        bad_tensor8 = aset['8']

        tf_dset = make_tf_dataset(aset, keys=keys)
        tf_dset = tf_dset.batch(3)
        total_batches = 0
        for dset1 in tf_dset:
            total_batches += 1
            assert dset1[0].shape == tf.TensorShape((3, 5, 7))
            for sample in dset1[0]:
                assert not np.allclose(sample, bad_tensor0)
                assert not np.allclose(sample, bad_tensor1)
                assert not np.allclose(sample, bad_tensor3)
                assert not np.allclose(sample, bad_tensor8)
        assert total_batches == 3
        co.close()

    @pytest.mark.filterwarnings("ignore:Dataloaders are experimental")
    def test_with_index_range(self, repo_20_filled_samples):
        repo = repo_20_filled_samples
        co = repo.checkout()
        aset = co.columns['writtenaset']

        # with keys
        bad_tensor0 = aset['0']
        bad_tensor1 = aset['1']

        # with index range
        index_range = slice(2, 20)
        tf_dset = make_tf_dataset(aset, index_range=index_range)
        tf_dset = tf_dset.batch(3)
        total_batches = 0
        for dset1 in tf_dset:
            total_batches += 1
            assert dset1[0].shape == tf.TensorShape((3, 5, 7))
            for sample in dset1[0]:
                assert not np.allclose(sample, bad_tensor0)
                assert not np.allclose(sample, bad_tensor1)
        assert total_batches == 6
        co.close()

    @pytest.mark.filterwarnings("ignore:Dataloaders are experimental")
    def test_dataset_loader_fails_with_write_enabled_checkout(self, repo_20_filled_samples):
        repo = repo_20_filled_samples
        co = repo.checkout(write=True)
        first_aset = co.columns['writtenaset']
        second_aset = co.columns['second_aset']
        with pytest.raises(TypeError):
            make_tf_dataset([first_aset, second_aset])
        co.close()

    @pytest.mark.filterwarnings("ignore:Dataloaders are experimental")
    def test_variably_shaped(self, aset_samples_var_shape_initialized_repo):
        # Variably shaped test is required since the collation is dependent on
        # the way we return the data from generator
        repo = aset_samples_var_shape_initialized_repo
        co = repo.checkout(write=True)
        aset = co.columns['writtenaset']
        for i in range(5, 10):
            aset[i] = np.random.random((2, i))
        co.commit('added data')
        co.close()

        co = repo.checkout()
        aset = co.columns['writtenaset']
        tf_dset = make_tf_dataset(aset)
        shape_obj = tf.TensorShape((2, None))
        tf_dset = tf_dset.padded_batch(5, padded_shapes=(shape_obj,))
        for val in tf_dset:
            assert val[0].shape[0] == 5
            assert val[0].shape[1] == 2
            assert 11 > val[0].shape[2] > 4
        co.close()

    @pytest.mark.filterwarnings("ignore:Dataloaders are experimental")
    def test_lots_of_data_with_multiple_backend(self, repo_300_filled_samples):
        repo = repo_300_filled_samples
        co = repo.checkout()
        aset = co.columns['aset']
        tf_dset = make_tf_dataset([aset])
        tf_dset = tf_dset.batch(10)
        for data in tf_dset:
            assert data[0].shape == (10, 5, 7)
        co.close()

    @pytest.mark.filterwarnings("ignore:Dataloaders are experimental")
    @pytest.mark.filterwarnings("ignore:Column.* writtenaset contains `reference-only` samples")
    def test_local_without_data_fails_no_common_no_local(self, written_two_cmt_server_repo, managed_tmpdir):
        new_tmpdir = pjoin(managed_tmpdir, 'new')
        mkdir(new_tmpdir)
        server, _ = written_two_cmt_server_repo
        repo = Repository(path=new_tmpdir, exists=False)
        repo.clone('name', 'a@b.c', server, remove_old=True)
        co = repo.checkout()
        aset = co.columns['writtenaset']
        with pytest.raises(ValueError):
            tf_dset = make_tf_dataset(aset)
        co.close()
        repo._env._close_environments()

    @pytest.mark.filterwarnings("ignore:Dataloaders are experimental")
    @pytest.mark.filterwarnings("ignore:Column.* writtenaset contains `reference-only` samples")
    def test_local_without_data_fails_no_common(self, written_two_cmt_server_repo, managed_tmpdir):
        new_tmpdir = pjoin(managed_tmpdir, 'new')
        mkdir(new_tmpdir)
        server, _ = written_two_cmt_server_repo
        repo = Repository(path=new_tmpdir, exists=False)
        repo.clone('name', 'a@b.c', server, remove_old=True)
        co = repo.checkout()
        aset = co.columns['writtenaset']
        with pytest.raises(KeyError):
            tf_dset = make_tf_dataset(aset, keys=['1', -1])
        co.close()
        repo._env._close_environments()

    @pytest.mark.filterwarnings("ignore:Dataloaders are experimental")
    @pytest.mark.filterwarnings("ignore:Column.* writtenaset contains `reference-only` samples")
    def test_local_without_data_fails_data_unavailable(self, written_two_cmt_server_repo, managed_tmpdir):
        new_tmpdir = pjoin(managed_tmpdir, 'new')
        mkdir(new_tmpdir)
        server, _ = written_two_cmt_server_repo
        repo = Repository(path=new_tmpdir, exists=False)
        repo.clone('name', 'a@b.c', server, remove_old=True)
        co = repo.checkout()
        aset = co.columns['writtenaset']
        with pytest.raises(FileNotFoundError):
            tf_dset = make_tf_dataset(aset, keys=['1', '2'])
        co.close()
        repo._env._close_environments()



# header files for spec containers

cdef class HDF5_01_DataHashSpec:

    cdef readonly str backend
    cdef readonly str uid
    cdef readonly str checksum
    cdef readonly str dataset
    cdef readonly int dataset_idx
    cdef readonly tuple shape


cdef class HDF5_00_DataHashSpec:

    cdef readonly str backend
    cdef readonly str uid
    cdef readonly str checksum
    cdef readonly str dataset
    cdef readonly int dataset_idx
    cdef readonly tuple shape


cdef class NUMPY_10_DataHashSpec:

    cdef readonly str backend
    cdef readonly str uid
    cdef readonly str checksum
    cdef readonly int collection_idx
    cdef readonly tuple shape


cdef class LMDB_30_DataHashSpec:

    cdef readonly str backend
    cdef readonly str uid
    cdef readonly str row_idx
    cdef readonly  str checksum


cdef class REMOTE_50_DataHashSpec:

    cdef readonly str backend
    cdef readonly str schema_hash



import pytest
import numpy as np

from conftest import fixed_shape_backend_params, variable_shape_backend_params

all_backend_params = list(set(fixed_shape_backend_params).union(set(variable_shape_backend_params)))


@pytest.mark.parametrize('backend1', all_backend_params)
@pytest.mark.parametrize('backend2', all_backend_params)
def test_nested_context_manager_does_not_close_all_open(repo, backend1, backend2):
    co = repo.checkout(write=True)
    fooaset = co.add_ndarray_column('foo', prototype=np.arange(10), backend=backend1)
    baraset = co.add_ndarray_column('bar', prototype=np.arange(10), backend=backend2, contains_subsamples=True)

    with co:
        assert co.columns._any_is_conman() is True
        assert fooaset._is_conman is True
        assert baraset._is_conman is True
        with fooaset as foo:
            assert co.columns._any_is_conman() is True
            assert foo._is_conman is True
            assert fooaset._is_conman is True
            assert baraset._is_conman is True
        assert co.columns._any_is_conman() is True
        assert fooaset._is_conman is True
        assert baraset._is_conman is True
    assert co.columns._any_is_conman() is False
    co.close()



from .recordstructs cimport ColumnSchemaKey, \
    FlatColumnDataKey, \
    NestedColumnDataKey, \
    DataRecordVal

from .recordstructs import ColumnSchemaKey, \
    FlatColumnDataKey, \
    NestedColumnDataKey, \
    DataRecordVal

import ast

__all__ = [
    'schema_record_count_start_range_key',
    'schema_db_key_from_column',
    'schema_db_range_key_from_column_unknown_layout',
    'schema_column_record_from_db_key',
    'schema_hash_record_db_val_from_spec',
    'schema_spec_from_db_val',
    'schema_hash_db_key_from_digest',
    'data_record_digest_val_from_db_val',
    'data_record_db_val_from_digest',
    'flat_data_column_record_start_range_key',
    'flat_data_db_key_from_names',
    'flat_data_record_from_db_key',
    'nested_data_column_record_start_range_key',
    'nested_data_db_key_from_names',
    'nested_data_record_from_db_key',
    'dynamic_layout_data_record_from_db_key',
    'dynamic_layout_data_record_db_start_range_key',
    'hash_schema_db_key_from_raw_key',
    'hash_data_db_key_from_raw_key',
    'hash_schema_raw_key_from_db_key',
    'hash_data_raw_key_from_db_key',
    'schema_record_db_val_from_digest',
]


# ----------------------- Schema Record Parsers -------------------------------


cpdef bytes schema_record_count_start_range_key():
    return 's:'.encode()


cpdef bytes schema_db_key_from_column(str column, str layout):
    """column schema db formated key from name and layout.

    Parameters
    ----------
    column: str
        name of the column
    layout: str
        layout of the column schema ('flat', 'nested', etc.)
    """
    cdef str serial

    if layout == 'flat':
        serial = f's:{column}:f'
    elif layout == 'nested':
        serial = f's:{column}:n'
    else:
        raise ValueError(f'layout {layout} not valid')
    return serial.encode()


cpdef bytes schema_db_range_key_from_column_unknown_layout(str column):
    """Find a cursor range key which will select a column schema key.
    
    Due to how information is appended onto the end of the schema db key,
    there is no need to know the column_layout or schema_digest to uniquely
    identify a column's schema record. set the cursor range and query the full
    key value (passed into a seperate parser) to recieve the column_layout 
    and schema_digest / hash type code. The schema spec is accessed at the 
    record value, or in the hash db under the corresponding schema_digest key.
    
    Parameters
    ----------
    column: str
        name of the column to query.
    """
    cdef str serial

    serial = f's:{column}:'
    return serial.encode()


cpdef ColumnSchemaKey schema_column_record_from_db_key(bytes raw):
    cdef str serial, column, layout

    serial = raw.decode()
    _, column, layout = serial.split(':')
    if layout == 'f':
        layout = 'flat'
    elif layout == 'n':
        layout = 'nested'
    else:
        raise ValueError(f'layout unknown for serial key {serial}')
    return ColumnSchemaKey(column, layout)


cpdef bytes schema_hash_record_db_val_from_spec(dict schema):
    cdef str serial

    serial = repr(schema).replace(' ', '')
    return serial.encode()


cpdef dict schema_spec_from_db_val(bytes raw):
    cdef str serialized
    cdef dict schema

    serialized = raw.decode()
    schema = ast.literal_eval(serialized)
    return schema


cpdef bytes schema_hash_db_key_from_digest(str digest):
    return f's:{digest}'.encode()


cpdef bytes schema_record_db_val_from_digest(str digest):
    return digest.encode()


# -------------------- Data Digest Record Value Parser -------------------------


cpdef DataRecordVal data_record_digest_val_from_db_val(bytes raw):
    """Convert and split a lmdb record value into data record val struct
    """
    cdef str serial

    serial = raw.decode()
    return DataRecordVal(serial)


cpdef bytes data_record_db_val_from_digest(str digest):
    """convert a data digest value spec into the appropriate lmdb record value
    """
    return digest.encode()


# -------------------------- flat parser --------------------------------------


cpdef bytes flat_data_column_record_start_range_key(str column):
    cdef str serial

    serial = f'f:{column}:'
    return serial.encode()


cpdef bytes flat_data_db_key_from_names(str column, sample):
    cdef str serial

    if isinstance(sample, int):
        serial = f'f:{column}:#{sample}'
    else:
        serial = f'f:{column}:{sample}'
    return serial.encode()


cpdef FlatColumnDataKey flat_data_record_from_db_key(bytes raw):
    cdef str serial, column, sample

    serial = raw.decode()
    _, column, sample = serial.split(':')
    return FlatColumnDataKey(column, sample)


# -------------------------- nested parser ------------------------------------


cpdef bytes nested_data_column_record_start_range_key(str column):
    cdef str serial

    serial = f'n:{column}:'
    return serial.encode()


cpdef bytes nested_data_db_key_from_names(str column, sample, subsample):
    cdef str serial

    if isinstance(sample, int):
        sample = f'#{sample}'
    if isinstance(subsample, int):
        subsample = f'#{subsample}'
    serial = f'n:{column}:{sample}:{subsample}'
    return serial.encode()


cpdef NestedColumnDataKey nested_data_record_from_db_key(bytes raw):
    cdef str serial, column, sample, subsample

    serial = raw.decode()
    _, column, sample, subsample = serial.split(':')
    return NestedColumnDataKey(column, sample, subsample)


# ----------------------- dynamic parser selection ----------------------------


cpdef object dynamic_layout_data_record_from_db_key(bytes raw):
    if raw[0:2] == b'f:':
        res = flat_data_record_from_db_key(raw)
    elif raw[0:2] == b'n:':
        res = nested_data_record_from_db_key(raw)
    elif raw[0:2] == b's:':
        res = schema_column_record_from_db_key(raw)
    else:
        raise ValueError(raw)
    return res


cpdef bytes dynamic_layout_data_record_db_start_range_key(ColumnSchemaKey column_record):
    cdef bytes res

    if column_record.layout == 'flat':
        res = flat_data_column_record_start_range_key(column_record.column)
    elif column_record.layout == 'nested':
        res = nested_data_column_record_start_range_key(column_record.column)
    else:
        raise ValueError(column_record)
    return res


#
# Data Hash parsing functions used to convert db key/val to raw pyhon obj
# -----------------------------------------------------------------------


cpdef bytes hash_record_count_start_range_key():
    return 'h:'.encode()


cpdef bytes hash_schema_db_key_from_raw_key(str schema_hash):
    return f's:{schema_hash}'.encode()


cpdef bytes hash_data_db_key_from_raw_key(str data_hash):
    return f'h:{data_hash}'.encode()


cpdef str hash_schema_raw_key_from_db_key(bytes db_key):
    return db_key[2:].decode()


cpdef str hash_data_raw_key_from_db_key(bytes db_key):
    return db_key[2:].decode()



from pathlib import Path
import weakref
import warnings
from typing import Union, Optional, List

from .merger import select_merge_algorithm
from .constants import DIR_HANGAR
from .remotes import Remotes
from .context import Environments
from .diagnostics import ecosystem, integrity
from .records import heads, parsing, summarize, vcompat, commiting
from .checkout import ReaderCheckout, WriterCheckout
from .diff import DiffAndConflicts, ReaderUserDiff
from .utils import (
    is_valid_directory_path,
    is_suitable_user_key,
    is_ascii,
    folder_size,
    format_bytes
)


class Repository(object):
    """Launching point for all user operations in a Hangar repository.

    All interaction, including the ability to initialize a repo, checkout a
    commit (for either reading or writing), create a branch, merge branches, or
    generally view the contents or state of the local repository starts here.
    Just provide this class instance with a path to an existing Hangar
    repository, or to a directory one should be initialized, and all required
    data for starting your work on the repo will automatically be populated.

        >>> from hangar import Repository
        >>> repo = Repository('foo/path/to/dir')

    Parameters
    ----------
    path : Union[str, os.PathLike]
        local directory path where the Hangar repository exists (or initialized)
    exists : bool, optional
        True if a Hangar repository should exist at the given directory path.
        Should no Hangar repository exists at that location, a UserWarning will
        be raised indicating that the :meth:`init` method needs to be called.

        False if the provided path does not need to (but optionally can) contain a
        Hangar repository.  if a Hangar repository does not exist at that path, the
        usual UserWarning will be suppressed.

        In both cases, the path must exist and the user must have sufficient OS
        permissions to write to that location. Default = True
    """

    def __init__(self, path: Union[str, Path], exists: bool = True):

        if isinstance(path, (str, bytes)):
            path = Path(path)

        try:
            usr_path = is_valid_directory_path(path)
        except (TypeError, NotADirectoryError, PermissionError) as e:
            raise e from None

        repo_pth = usr_path.joinpath(DIR_HANGAR)
        if exists is False:
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                envs = Environments(pth=repo_pth)
        else:
            envs = Environments(pth=repo_pth)

        self._repo_path: Path = repo_pth
        self._env: Environments = envs
        self._remote: Remotes = Remotes(self._env)

    def _repr_pretty_(self, p, cycle):
        """provide a pretty-printed repr for ipython based user interaction.

        Parameters
        ----------
        p : printer
            io stream printer type object which is provided via ipython
        cycle : bool
            if the pretty-printer detects a cycle or infinite loop. Not a
            concern here since we just output the text and return, no looping
            required.

        """
        self.__verify_repo_initialized()
        res = f'Hangar {self.__class__.__name__}\
               \n    Repository Path  : {self.path}\
               \n    Writer-Lock Free : {heads.writer_lock_held(self._env.branchenv)}\n'
        p.text(res)

    def __repr__(self):
        """Override the default repr to show useful information to developers.

        Note: the pprint repr (ipython enabled) is separately defined in
        :py:meth:`_repr_pretty_`. We specialize because we assume that anyone
        operating in a terminal-based interpreter is probably a more advanced
        developer-type, and expects traditional repr information instead of a
        user facing summary of the repo. Though if we're wrong, go ahead and
        feel free to reassign the attribute :) won't hurt our feelings, promise.

        Returns
        -------
        string
            formatted representation of the object
        """
        res = f'{self.__class__}(path={self._repo_path})'
        return res

    def __verify_repo_initialized(self):
        """Internal method to verify repo initialized before operations occur

        Raises
        ------
        RuntimeError
            If the repository db environments have not been initialized at the
            specified repo path.
        """
        if not self._env.repo_is_initialized:
            msg = f'Repository at path: {self._repo_path} has not been initialized. '\
                  f'Please run the `init_repo()` function'
            raise RuntimeError(msg)

    @property
    def remote(self) -> Remotes:
        """Accessor to the methods controlling remote interactions.

        .. seealso::

           :class:`Remotes` for available methods of this property

        Returns
        -------
        Remotes
            Accessor object methods for controlling remote interactions.
        """
        proxy = weakref.proxy(self._remote)
        return proxy

    @property
    def path(self) -> str:
        """Return the path to the repository on disk, read-only attribute

        Returns
        -------
        str
            path to the specified repository, not including `.hangar` directory
        """
        self.__verify_repo_initialized()
        return str(self._repo_path.parent)

    @property
    def writer_lock_held(self) -> bool:
        """Check if the writer lock is currently marked as held. Read-only attribute.

        Returns
        -------
        bool
            True is writer-lock is held, False if writer-lock is free.
        """
        self.__verify_repo_initialized()
        return not heads.writer_lock_held(self._env.branchenv)

    @property
    def version(self) -> str:
        """Find the version of Hangar software the repository is written with

        Returns
        -------
        str
            semantic version of major, minor, micro version of repo software version.
        """
        self.__verify_repo_initialized()
        res = vcompat.get_repository_software_version_spec(self._env.branchenv)
        return str(res)

    @property
    def initialized(self) -> bool:
        """
        Check if the repository has been initialized or not

        Returns
        -------
        bool
            True if repository has been initialized.
        """
        return self._env.repo_is_initialized

    @property
    def size_nbytes(self) -> int:
        """Disk space used by the repository returned in number of bytes.

            >>> repo.size_nbytes
            1234567890
            >>> print(type(repo.size_nbytes))
            <class 'int'>

        Returns
        -------
        int
            number of bytes used by the repository on disk.
        """
        self.__verify_repo_initialized()
        return folder_size(self._repo_path, recurse=True)

    @property
    def size_human(self) -> str:
        """Disk space used by the repository returned in human readable string.

            >>> repo.size_human
            '1.23 GB'
            >>> print(type(repo.size_human))
            <class 'str'>

        Returns
        -------
        str
            disk space used by the repository formated in human readable text.
        """
        self.__verify_repo_initialized()
        nbytes = folder_size(self._repo_path, recurse=True)
        return format_bytes(nbytes)

    def checkout(self,
                 write: bool = False,
                 *,
                 branch: str = '',
                 commit: str = '') -> Union[ReaderCheckout, WriterCheckout]:
        """Checkout the repo at some point in time in either `read` or `write` mode.

        Only one writer instance can exist at a time. Write enabled checkout
        must must create a staging area from the ``HEAD`` commit of a branch. On
        the contrary, any number of reader checkouts can exist at the same time
        and can specify either a branch name or a commit hash.

        Parameters
        ----------
        write : bool, optional
            Specify if the checkout is write capable, defaults to False
        branch : str, optional
            name of the branch to checkout. This utilizes the state of the repo
            as it existed at the branch ``HEAD`` commit when this checkout object
            was instantiated, defaults to ''
        commit : str, optional
            specific hash of a commit to use for the checkout (instead of a
            branch ``HEAD`` commit). This argument takes precedent over a branch
            name parameter if it is set. Note: this only will be used in
            non-writeable checkouts, defaults to ''

        Raises
        ------
        ValueError
            If the value of `write` argument is not boolean
        ValueError
            If ``commit`` argument is set to any value when ``write=True``.
            Only ``branch`` argument is allowed.

        Returns
        -------
        Union[ReaderCheckout, WriterCheckout]
            Checkout object which can be used to interact with the repository
            data
        """
        self.__verify_repo_initialized()
        try:
            if write is True:
                if commit != '':
                    raise ValueError(
                        f'Only `branch` argument can be set if `write=True`. '
                        f'Setting `commit={commit}` not allowed.')
                if branch == '':
                    branch = heads.get_staging_branch_head(self._env.branchenv)
                co = WriterCheckout(
                    repo_pth=self._repo_path,
                    branch_name=branch,
                    hashenv=self._env.hashenv,
                    refenv=self._env.refenv,
                    stageenv=self._env.stageenv,
                    branchenv=self._env.branchenv,
                    stagehashenv=self._env.stagehashenv)
                return co
            elif write is False:
                commit_hash = self._env.checkout_commit(
                    branch_name=branch, commit=commit)
                co = ReaderCheckout(
                    base_path=self._repo_path,
                    dataenv=self._env.cmtenv[commit_hash],
                    hashenv=self._env.hashenv,
                    branchenv=self._env.branchenv,
                    refenv=self._env.refenv,
                    commit=commit_hash)
                return co
            else:
                raise ValueError("Argument `write` only takes True or False as value")
        except (RuntimeError, ValueError) as e:
            raise e from None

    def clone(self, user_name: str, user_email: str, remote_address: str,
              *, remove_old: bool = False) -> str:
        """Download a remote repository to the local disk.

        The clone method implemented here is very similar to a `git clone`
        operation. This method will pull all commit records, history, and data
        which are parents of the remote's `master` branch head commit. If a
        :class:`Repository` exists at the specified directory,
        the operation will fail.

        Parameters
        ----------
        user_name : str
            Name of the person who will make commits to the repository. This
            information is recorded permanently in the commit records.
        user_email : str
            Email address of the repository user. This information is recorded
            permanently in any commits created.
        remote_address : str
            location where the
            :class:`hangar.remote.server.HangarServer` process is
            running and accessible by the clone user.
        remove_old : bool, optional, kwarg only
            DANGER! DEVELOPMENT USE ONLY! If enabled, a
            :class:`hangar.repository.Repository` existing on disk at the same
            path as the requested clone location will be completely removed and
            replaced with the newly cloned repo. (the default is False, which
            will not modify any contents on disk and which will refuse to create
            a repository at a given location if one already exists there.)

        Returns
        -------
        str
            Name of the master branch for the newly cloned repository.
        """
        self.init(user_name=user_name, user_email=user_email, remove_old=remove_old)
        self._remote.add(name='origin', address=remote_address)
        branch = self._remote.fetch(remote='origin', branch='master')
        HEAD = heads.get_branch_head_commit(self._env.branchenv, branch_name=branch)
        heads.set_branch_head_commit(self._env.branchenv, 'master', HEAD)
        with warnings.catch_warnings(record=False):
            warnings.simplefilter('ignore', category=UserWarning)
            co = self.checkout(write=True, branch='master')
            co.reset_staging_area()
            co.close()
        return 'master'

    def init(self,
             user_name: str,
             user_email: str,
             *,
             remove_old: bool = False) -> str:
        """Initialize a Hangar repository at the specified directory path.

        This function must be called before a checkout can be performed.

        Parameters
        ----------
        user_name : str
            Name of the repository user account.
        user_email : str
            Email address of the repository user account.
        remove_old : bool, kwarg-only
            DEVELOPER USE ONLY -- remove and reinitialize a Hangar
            repository at the given path, Default = False

        Returns
        -------
        str
            the full directory path where the Hangar repository was
            initialized on disk.
        """
        pth = self._env.init_repo(user_name=user_name,
                                  user_email=user_email,
                                  remove_old=remove_old)
        return str(pth)

    def log(self,
            branch: str = None,
            commit: str = None,
            *,
            return_contents: bool = False,
            show_time: bool = False,
            show_user: bool = False) -> Optional[dict]:
        """Displays a pretty printed commit log graph to the terminal.

        .. note::

            For programatic access, the return_contents value can be set to true
            which will retrieve relevant commit specifications as dictionary
            elements.

        Parameters
        ----------
        branch : str, optional
            The name of the branch to start the log process from. (Default value
            = None)
        commit : str, optional
            The commit hash to start the log process from. (Default value = None)
        return_contents : bool, optional, kwarg only
            If true, return the commit graph specifications in a dictionary
            suitable for programatic access/evaluation.
        show_time : bool, optional, kwarg only
            If true and return_contents is False, show the time of each commit
            on the printed log graph
        show_user : bool, optional, kwarg only
            If true and return_contents is False, show the committer of each
            commit on the printed log graph
        Returns
        -------
        Optional[dict]
            Dict containing the commit ancestor graph, and all specifications.
        """
        self.__verify_repo_initialized()
        res = summarize.log(branchenv=self._env.branchenv,
                            refenv=self._env.refenv,
                            branch=branch,
                            commit=commit,
                            return_contents=return_contents,
                            show_time=show_time,
                            show_user=show_user)
        return res

    def summary(self, *, branch: str = '', commit: str = '') -> None:
        """Print a summary of the repository contents to the terminal

        Parameters
        ----------
        branch : str, optional
            A specific branch name whose head commit will be used as the summary
            point (Default value = '')
        commit : str, optional
            A specific commit hash which should be used as the summary point.
            (Default value = '')
        """
        self.__verify_repo_initialized()
        ppbuf = summarize.summary(self._env, branch=branch, commit=commit)
        print(ppbuf.getvalue())
        return None

    def _details(self, *, line_limit=100, line_length=100) -> None:  # pragma: no cover
        """DEVELOPER USE ONLY: Dump some details about the underlying db structure to disk.
        """
        print(summarize.details(
            self._env.branchenv, line_limit=line_limit, line_length=line_length).getvalue())
        print(summarize.details(
            self._env.refenv, line_limit=line_limit, line_length=line_length).getvalue())
        print(summarize.details(
            self._env.hashenv, line_limit=line_limit, line_length=line_length).getvalue())
        print(summarize.details(
            self._env.stageenv, line_limit=line_limit, line_length=line_length).getvalue())
        print(summarize.details(
            self._env.stagehashenv, line_limit=line_limit, line_length=line_length).getvalue())
        for commit, commitenv in self._env.cmtenv.items():
            print(summarize.details(
                commitenv, line_limit=line_limit, line_length=line_length).getvalue())
        return

    def _ecosystem_details(self) -> dict:
        """DEVELOPER USER ONLY: log and return package versions on the system.
        """
        eco = ecosystem.get_versions()
        return eco

    def diff(self, master: str, dev: str) -> DiffAndConflicts:
        """Calculate diff between master and dev branch/commits.

        Diff is calculated as if we are to merge "dev" into "master"

        Parameters
        ----------
        master: str
            branch name or commit hash digest to use as the "master" which
            changes made in "dev" are compared to.
        dev: str
            branch name or commit hash digest to use as the "dev"
            (ie. "feature") branch which changes have been made to
            which are to be compared to the contents of "master".

        Returns
        -------
        DiffAndConflicts
            Standard output diff structure.
        """
        current_branches = self.list_branches()

        # assert branch / commit specified by "master" exists and
        # standardize into "digest" rather than "branch name" arg type
        if master in current_branches:
            masterHEAD = heads.get_branch_head_commit(
                branchenv=self._env.branchenv, branch_name=master)
        else:
            cmtExists = commiting.check_commit_hash_in_history(
                refenv=self._env.refenv, commit_hash=master)
            if not cmtExists:
                raise ValueError(f'`master` {master} is not valid branch/commit.')
            masterHEAD = master

        # same check & transform for "dev" branch/commit arg.
        if dev in current_branches:
            devHEAD = heads.get_branch_head_commit(
                branchenv=self._env.branchenv, branch_name=dev)
        else:
            cmtExists = commiting.check_commit_hash_in_history(
                refenv=self._env.refenv, commit_hash=dev)
            if not cmtExists:
                raise ValueError(f'`dev` {dev} is not valid branch/commit.')
            devHEAD = dev

        # create differ object and generate results...
        diff = ReaderUserDiff(commit_hash=masterHEAD,
                              branchenv=self._env.branchenv,
                              refenv=self._env.refenv)
        res = diff.commit(dev_commit_hash=devHEAD)
        return res


    def merge(self, message: str, master_branch: str, dev_branch: str) -> str:
        """Perform a merge of the changes made on two branches.

        Parameters
        ----------
        message: str
            Commit message to use for this merge.
        master_branch : str
            name of the master branch to merge into
        dev_branch : str
            name of the dev/feature branch to merge

        Returns
        -------
        str
            Hash of the commit which is written if possible.
        """
        self.__verify_repo_initialized()
        commit_hash = select_merge_algorithm(
            message=message,
            branchenv=self._env.branchenv,
            stageenv=self._env.stageenv,
            refenv=self._env.refenv,
            stagehashenv=self._env.stagehashenv,
            master_branch=master_branch,
            dev_branch=dev_branch,
            repo_path=self._repo_path)

        return commit_hash

    def create_branch(self, name: str, base_commit: str = None) -> heads.BranchHead:
        """create a branch with the provided name from a certain commit.

        If no base commit hash is specified, the current writer branch ``HEAD``
        commit is used as the ``base_commit`` hash for the branch. Note that
        creating a branch does not actually create a checkout object for
        interaction with the data. to interact you must use the repository
        checkout method to properly initialize a read (or write) enabled
        checkout object.

            >>> from hangar import Repository
            >>> repo = Repository('foo/path/to/dir')

            >>> repo.create_branch('testbranch')
                BranchHead(name='testbranch', digest='b66b...a8cc')
            >>> repo.list_branches()
                ['master', 'testbranch']
            >>> co = repo.checkout(write=True, branch='testbranch')
            >>> # add data ...
            >>> newDigest = co.commit('added some stuff')

            >>> repo.create_branch('new-changes', base_commit=newDigest)
                BranchHead(name='new-changes', digest='35kd...3254')
            >>> repo.list_branches()
                ['master', 'new-changes', 'testbranch']

        Parameters
        ----------
        name : str
            name to assign to the new branch
        base_commit : str, optional
            commit hash to start the branch root at. if not specified, the
            writer branch ``HEAD`` commit at the time of execution will be used,
            defaults to None

        Returns
        -------
        :class:`~.heads.BranchHead`
            NamedTuple[str, str] with fields for ``name`` and ``digest`` of the
            branch created (if the operation was successful)

        Raises
        ------
        ValueError
            If the branch name provided contains characters outside of alpha-numeric
            ascii characters and ".", "_", "-" (no whitespace), or is > 64 characters.
        ValueError
            If the branch already exists.
        RuntimeError
            If the repository does not have at-least one commit on the "default"
            (ie. ``master``) branch.
        """
        self.__verify_repo_initialized()
        if (not is_ascii(name)) or (not is_suitable_user_key(name)):
            err = ValueError(
                f'Branch name provided: {name} invalid. Must contain only alpha-numeric '
                f'or "." "_" "-" ascii characters. And be <= 64 Characters')
            raise err from None
        createdBranch = heads.create_branch(
            branchenv=self._env.branchenv,
            name=name,
            base_commit=base_commit)
        return createdBranch

    def remove_branch(self, name: str, *, force_delete: bool = False) -> heads.BranchHead:
        """Permanently delete a branch pointer from the repository history.

        Since a branch (by definition) is the name associated with the HEAD
        commit of a historical path, the default behavior of this method is to
        throw an exception (no-op) should the ``HEAD`` not be referenced as an
        ancestor (or at least as a twin) of a separate branch which is
        currently *ALIVE*. If referenced in another branch's history, we are
        assured that all changes have been merged and recorded, and that this
        pointer can be safely deleted without risk of damage to historical
        provenance or (eventual) loss to garbage collection.

            >>> from hangar import Repository
            >>> repo = Repository('foo/path/to/dir')

            >>> repo.create_branch('first-testbranch')
            BranchHead(name='first-testbranch', digest='9785...56da')
            >>> repo.create_branch('second-testbranch')
            BranchHead(name='second-testbranch', digest='9785...56da')
            >>> repo.list_branches()
            ['master', 'first-testbranch', 'second-testbranch']
            >>> # Make a commit to advance a branch
            >>> co = repo.checkout(write=True, branch='first-testbranch')
            >>> # add data ...
            >>> co.commit('added some stuff')
            '3l253la5hna3k3a553256nak35hq5q534kq35532'
            >>> co.close()

            >>> repo.remove_branch('second-testbranch')
            BranchHead(name='second-testbranch', digest='9785...56da')

        A user may manually specify to delete an un-merged branch, in which
        case the ``force_delete`` keyword-only argument should be set to
        ``True``.

            >>> # check out master and try to remove 'first-testbranch'
            >>> co = repo.checkout(write=True, branch='master')
            >>> co.close()

            >>> repo.remove_branch('first-testbranch')
            Traceback (most recent call last):
                ...
            RuntimeError: ("The branch first-testbranch is not fully merged. "
            "If you are sure you want to delete it, re-run with "
            "force-remove parameter set.")
            >>> # Now set the `force_delete` parameter
            >>> repo.remove_branch('first-testbranch', force_delete=True)
            BranchHead(name='first-testbranch', digest='9785...56da')

        It is important to note that *while this method will handle all safety
        checks, argument validation, and performs the operation to permanently
        delete a branch name/digest pointer, **no commit refs along the history
        will be deleted from the Hangar database**.* Most of the history contains
        commit refs which must be safe in other branch histories, and recent
        commits may have been used as the base for some new history. As such, even
        if some of the latest commits leading up to a deleted branch ``HEAD`` are
        orphaned (unreachable), the records (and all data added in those commits)
        will remain on the disk.

        In the future, we intend to implement a garbage collector which will remove
        orphan commits which have not been modified for some set amount of time
        (probably on the order of a few months), but this is not implemented at the
        moment.

        Should an accidental forced branch deletion occur, *it is possible to
        recover* and create a new branch head pointing to the same commit. If
        the commit digest of the removed branch ``HEAD`` is known, its as simple as
        specifying a name and the ``base_digest`` in the normal
        :meth:`create_branch` method. If the digest is unknown, it will be a
        bit more work, but some of the developer facing introspection tools /
        routines could be used to either manually or (with minimal effort)
        programmatically find the orphan commit candidates. If you find
        yourself having accidentally deleted a branch, and must get it back,
        please reach out on the `Github Issues
        <https://github.com/tensorwerk/hangar-py/issues>`__ page. We'll gladly
        explain more in depth and walk you through the process in any way we
        can help!

        Parameters
        ----------
        name : str
            name of the branch which should be deleted. This branch must exist, and
            cannot refer to a remote tracked branch (ie. origin/devbranch), please
            see exception descriptions for other parameters determining validity of
            argument
        force_delete : bool, optional
            If True, remove the branch pointer even if the changes are un-merged in
            other branch histories. May result in orphaned commits which may be
            time-consuming to recover if needed, by default False

        Returns
        -------
        :class:`~.heads.BranchHead`
            NamedTuple[str, str] with fields for `name` and `digest` of the branch
            pointer deleted.

        Raises
        ------
        ValueError
            If a branch with the provided name does not exist locally
        PermissionError
            If removal of the branch would result in a repository with zero local
            branches.
        PermissionError
            If a write enabled checkout is holding the writer-lock at time of this
            call.
        PermissionError
            If the branch to be removed was the last used in a write-enabled
            checkout, and whose contents form the base of the staging area.
        RuntimeError
            If the branch has not been fully merged into other branch histories,
            and ``force_delete`` option is not ``True``.
        """
        self.__verify_repo_initialized()
        res = heads.remove_branch(branchenv=self._env.branchenv,
                                  refenv=self._env.refenv,
                                  name=name,
                                  force_delete=force_delete)
        return res

    def list_branches(self) -> List[str]:
        """list all branch names created in the repository.

        Returns
        -------
        List[str]
            the branch names recorded in the repository
        """
        self.__verify_repo_initialized()
        branches = heads.get_branch_names(self._env.branchenv)
        return branches

    def verify_repo_integrity(self) -> bool:
        """Verify the integrity of the repository data on disk.

        Runs a full cryptographic verification of repository contents in order
        to ensure the integrity of all data and history recorded on disk.

        .. note::

            This proof may take a significant amount of time to run for
            repositories which:

            1. store significant quantities of data on disk.
            2. have a very large number of commits in their history.

            As a brief explanation for why these are the driving factors behind
            processing time:

            1. Every single piece of data in the repositories history must be read
               from disk, cryptographically hashed, and compared to the expected
               value. There is no exception to this rule; regardless of when a piece
               of data was added / removed from an column, or for how many (or how
               few) commits some sample exists in. The integrity of the commit tree at
               any point after some piece of data is added to the repo can only be
               validated if it - and all earlier data pieces - are proven to be intact
               and unchanged.

               Note: This does not mean that the verification is repeatedly
               performed for every commit some piece of data is stored in. Each
               data piece is read from disk and verified only once, regardless of
               how many commits some piece of data is referenced in.

            2. Each commit reference (defining names / contents of a commit) must be
               decompressed and parsed into a usable data structure. We scan across
               all data digests referenced in the commit and ensure that the
               corresponding data piece is known to hangar (and validated as
               unchanged). The commit refs (along with the corresponding user records,
               message, and parent map), are then re-serialized and cryptographically
               hashed for comparison to the expected value. While this process is
               fairly efficient for a single commit, it must be repeated for each
               commit in the repository history, and may take a non-trivial amount of
               time for repositories with thousands of commits.

        While the two points above are the most time consuming operations,
        there are many more checks which are performed alongside them as part
        of the full verification run.

        Returns
        -------
        bool
            True if integrity verification is successful, otherwise False; in
            this case, a message describing the offending component will be
            printed to stdout.
        """
        self.__verify_repo_initialized()
        heads.acquire_writer_lock(self._env.branchenv, 'VERIFY_PROCESS')
        try:
            integrity.run_verification(
                branchenv=self._env.branchenv,
                hashenv=self._env.hashenv,
                refenv=self._env.refenv,
                repo_path=self._env.repo_path)
        finally:
            heads.release_writer_lock(self._env.branchenv, 'VERIFY_PROCESS')
        return True

    def force_release_writer_lock(self) -> bool:
        """Force release the lock left behind by an unclosed writer-checkout

        .. warning::

            *NEVER USE THIS METHOD IF WRITER PROCESS IS CURRENTLY ACTIVE.* At the time
            of writing, the implications of improper/malicious use of this are not
            understood, and there is a a risk of of undefined behavior or (potentially)
            data corruption.

            At the moment, the responsibility to close a write-enabled checkout is
            placed entirely on the user. If the `close()` method is not called
            before the program terminates, a new checkout with write=True will fail.
            The lock can only be released via a call to this method.

        .. note::

            This entire mechanism is subject to review/replacement in the future.

        Returns
        -------
        bool
            if the operation was successful.
        """
        self.__verify_repo_initialized()
        forceReleaseSentinal = parsing.repo_writer_lock_force_release_sentinal()
        success = heads.release_writer_lock(self._env.branchenv, forceReleaseSentinal)
        return success
